from abc import ABCMeta, abstractmethod


class OutputInterceptionDataHandler(object):
    """
    A class that act as a pluggable hook that can be used during output interception when recording and playing the data
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def prepare_output_for_recording(self, interception_key, args, kwargs):
        """
        Prepare the input result that should be saved in the recording
        :param interception_key: Output interception key
        :type interception_key: basestring
        :param args: Output invocation args
        :type args: tuple
        :param kwargs: Output invocation kwargs
        :type kwargs: dict
        :return: Output result in a form that should be saved in the recording
        :rtype: Any
        """
        pass

    @abstractmethod
    def restore_output_from_recording(self, recorded_data):
        """
        Restore the actual input from the recording
        :param recorded_data: Recorded data provided by the prepare method
        :type recorded_data: Any
        :return: Object representing the output that was saved to the recording
        :rtype: Any
        """
        pass



import uuid
from abc import ABCMeta, abstractmethod


class Recording(object):
    """
    Holds a recording of an operation that was recorded using the TapeRecorder
    """

    __metaclass__ = ABCMeta

    def __init__(self, _id=None):
        self.id = _id or uuid.uuid1().hex
        self._closed = False

    @abstractmethod
    def _set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        pass

    def set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        assert not self._closed
        self._set_data(key, value)

    @abstractmethod
    def get_data(self, key):
        """
        Should always return a fresh copy of the data event when called multiple times with the same key,
        to prevent modifications of the recording by the calling code. These modifications can lead to differences in
        outputs that are not expected and will yield false positives.
        :param key: Data key
        :type key: basestring
        :return: Recorded data under given key
        :rtype: Any
        :raise: playback.exceptions.RecordingKeyError
        """
        pass

    @abstractmethod
    def get_data_direct(self, key):
        """
        Get the data directly from the recording with necessarily copying it, this should be used with care for
        performance reasons, mainly to be accessed during the recording stage as it can return the underlying saved
        data and not a copy
        :param key: Data key
        :type key: basestring
        :return: Recorded data under given key
        :rtype: Any
        :raise: playback.exceptions.RecordingKeyError
        """
        pass

    @abstractmethod
    def get_all_keys(self):
        """
        :return: All recorded keys
        :rtype: list of basestring
        """
        pass

    def add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        assert not self._closed
        self._add_metadata(metadata)

    @abstractmethod
    def _add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        pass

    def close(self):
        """
        Close the recording and release any underlying resources
        """
        self._closed = True

    def __setitem__(self, key, value):
        self._set_data(key, value)

    def __getitem__(self, item):
        return self.get_data(item)

    @abstractmethod
    def get_metadata(self):
        """
        :return: Recorded metadata
        :rtype: dict
        """
        pass



import multiprocessing as mp
import os
import signal
import sys
import traceback
from collections import Counter, namedtuple
from time import time

from enum import Enum

try:
    from StringIO import StringIO  # SKIP_FUTURIZE for Python 2
except ImportError:
    from io import StringIO  # for Python 3

import logging

_logger = logging.getLogger(__name__)


"""
Equal               - The recorded and playback compared outputs are considered equal
Fixed               - The playback output is considered as a fix to the matched errornous recorded output
Different           - The recorded and playback compared outputs are considered different
Failed              - The playback output is considered as a failure compared to the recorded output
EqualizerFailure    - The framework have failed running the comparison
"""
EqualityStatus = Enum('EqualityStatus', 'Equal Fixed Different Failed EqualizerFailure')


class ComparatorResult(object):
    def __init__(self, equality_status, message=None, diff=None):
        """
        :param equality_status: Equality status between expected and actual
        :type equality_status: EqualityStatus
        :param message: Optional comparison message
        :type message: basestring
        """
        self.equality_status = equality_status
        self.message = message
        self.diff = diff

    def __str__(self):
        if not self.message:
            return u'{}'.format(self.equality_status.name)

        return u'{} - {}'.format(self.equality_status.name, self.message)

    @staticmethod
    def failure_result(recording_id, exception):
        """
        :param recording_id: Id of the recording that is being played
        :type recording_id: str
        :param exception: Exception to wrap with failure result
        :type exception: builtins.Exception
        :return: Failure comparator result representing the given exception
        :rtype: ComparatorResult
        """
        _logger.warning(u'Failed playing recording id {} - {}'.format(recording_id, exception))
        trace_str = Equalizer.get_exception_stack_trace()
        _logger.warning(trace_str)
        return ComparatorResult(EqualityStatus.EqualizerFailure, str(exception))


class Comparison(object):
    def __init__(self, comparator_result, expected, actual, expected_is_exception, actual_is_exception, playback,
                 recording_id):
        """
        :param comparator_result: Result of comparison between expected and actual
        :type comparator_result: ComparatorResult
        :param expected: Expected result (recording)
        :type expected: Any
        :param actual: Actual result (playback)
        :type actual: Any
        :param expected_is_exception: Is expected result an exception (recording)
        :type expected_is_exception: bool
        :param actual_is_exception: Is actual result an exception (recording)
        :type actual_is_exception: bool
        :param playback: Play operation result
        :type playback: playback.tape_recorder.Playback
        :param recording_id: Id of compared recording
        :type recording_id: str
        """
        self.comparator_status = comparator_result
        self.expected = expected
        self.actual = actual
        self.expected_is_exception = expected_is_exception
        self.actual_is_exception = actual_is_exception
        self.playback = playback
        self.recording_id = recording_id

    def __str__(self):
        result = self.comparator_status.equality_status.name
        if self.comparator_status.message:
            result += u' - ' + self.comparator_status.message
        return result


class CompareExecutionConfig(object):
    def __init__(self, keep_results_in_comparison=False, compare_in_dedicated_process=False,
                 compare_process_recycle_rate=5, compare_process_timeout=10 * 60):
        """
        :param keep_results_in_comparison: Whether to keep results in the comparison result object
        :type keep_results_in_comparison: bool
        :param compare_in_dedicated_process: Whether to conduct the comparison in a dedicated process that is recycled
        every X comparisons or conduct in this process main thread
        :type compare_in_dedicated_process: bool
        :param compare_process_recycle_rate: If comparing in a dedicated process, specified after how many comparison
        should that process be recycled
        :type compare_process_recycle_rate: int
        :param compare_process_timeout: Time in seconds to wait for compare process to end before aborting it
        :type compare_process_timeout: float
        """
        self.keep_results_in_comparison = keep_results_in_comparison
        self.compare_in_dedicated_process = compare_in_dedicated_process
        self.compare_process_recycle_rate = compare_process_recycle_rate
        self.compare_process_timeout = compare_process_timeout


PlayAndCompareResult = namedtuple(
        'PlayAndCompareResult',
        'comparator_result playback recorded_result_is_exception playback_result_is_exception')


class Equalizer(object):
    def __init__(self, recording_ids, player, result_extractor, comparator, comparison_data_extractor=None,
                 compare_execution_config=None):
        """
        :param recording_ids: Iterator of playable recordings to compare
        :type recording_ids: collections.Iterator[str]
        :param player: A function that plays a recording given an id
        :type player: function
        collections.Iterator[playback.comparison.recordings_lookup.PlayableRecording]
        :param result_extractor: A function used to extract the results that needs to be compared from the recording
        and playback outputs
        :type result_extractor: function
        :param comparison_data_extractor: A function used to extract optional data from the recording that will be
        passed to the comparator
        :type comparison_data_extractor: function
        :param comparator: A function used to create the comparison result by comparing the recorded vs replayed result
        :type comparator: function
        :param compare_execution_config: Configuration specific to the comparison execution flow
        :type compare_execution_config: CompareExecutionConfig
        """
        self.recording_ids = recording_ids
        self.player = player
        self.result_extractor = result_extractor
        self.comparison_data_extractor = comparison_data_extractor
        self.comparator = comparator
        self.compare_execution_config = compare_execution_config or CompareExecutionConfig()
        # Init multiprocess related properties
        self._compare_tasks = mp.Queue()
        self._compare_results = mp.Queue()
        self._terminate_process = mp.Event()
        self._compare_process = None
        self._compare_process_age = 0

    def run_comparison(self):
        """
        Runs a comparison between recorded results and their corresponding playbacks
        :return: Comparison result
        :rtype: collections.Iterator[Comparison]
        """
        counter = Counter()
        iteration = 0
        completed = False
        try:
            for iteration, recording_id in enumerate(self.recording_ids, start=1):
                try:
                    play_and_compare_result = self._play_and_compare_recording_within_worker(recording_id)
                    playback = play_and_compare_result.playback

                    # Since comparison was done in another process and result may not be serializable accross processes,
                    # we have to re-extract here again if needed, if playback is None it means we had an error too early
                    # to be able to extract results
                    if playback is not None and self.compare_execution_config.keep_results_in_comparison:
                        recorded_result = self.result_extractor(playback.recorded_outputs)
                        playback_result = self.result_extractor(playback.playback_outputs)
                    else:
                        recorded_result = None
                        playback_result = None

                    comparison = Comparison(
                        play_and_compare_result.comparator_result,
                        recorded_result,
                        playback_result,
                        play_and_compare_result.recorded_result_is_exception,
                        play_and_compare_result.playback_result_is_exception,
                        playback,
                        recording_id
                    )

                    _logger.info(u'Recording {} Comparison result: {}'.format(recording_id,
                                                                              comparison))

                    counter[comparison.comparator_status.equality_status] += 1

                    if iteration % 10 == 0:
                        _logger.info(u'Iteration {} {}'.format(
                            iteration, Equalizer._comparison_stats_repr(counter)))

                    yield comparison
                except Exception as ex:  # pylint: disable=broad-except
                    counter[EqualityStatus.EqualizerFailure] += 1
                    yield Comparison(
                        ComparatorResult.failure_result(recording_id, ex),
                        None,
                        None,
                        False,
                        False,
                        None,
                        recording_id)

            completed = True

        finally:
            self._terminate_process.set()
            self._compare_tasks.close()
            self._compare_results.close()
            log_prefix = u'Completed all' if completed else u'Error during playback, executed'
            _logger.info(u'{} {} iterations, {}'.format(
                log_prefix, iteration, Equalizer._comparison_stats_repr(counter)))

    def _play_and_compare_recording_within_worker(self, recording_id):
        """
        Play the given recording id and compare the outputs in a worker, the worker process is determined by the
        'playback_in_dedicated_process' flag, If True it will use a dedicated process, else it will run within this
        process
        :param recording_id: Recording id to play
        :type recording_id: str
        :return: Comparison result, playback result, is recorded result an exception,
        is playback result an exception
        :rtype: PlayAndCompareResult
        """
        if not self.compare_execution_config.compare_in_dedicated_process:
            return self._play_and_compare_recording(recording_id)

        self._create_or_recycle_player_process_if_needed()

        # Queue the task for the playback process and wait for its result
        self._compare_tasks.put(recording_id)
        start_time = time()
        timed_out = True
        while time() - start_time <= self.compare_execution_config.compare_process_timeout:
            try:
                succeeded, result = self._compare_results.get(True, 1)
                timed_out = False

                if not succeeded:
                    raise Exception(result)

                break
            except mp.queues.Empty:
                if not self._compare_process.is_alive():
                    self._compare_process = None
                    raise Exception("playback process have died")

        if timed_out:
            self._handle_compare_execution_timeout()

        return result

    def _handle_compare_execution_timeout(self):
        """
        Handle the case that we had a timeout during comparison, killing the process if it is still alive
        """
        _logger.warning('Waiting for comparison result timed out')
        if self._compare_process.is_alive():
            try:
                self._kill_compare_process()
            except OSError as ex:
                # Don't fail when could not kill
                _logger.warning(u'Error while killing worker, {}'.format(str(ex)))
        self._compare_process = None
        raise Exception("timeout while running recording playback and comparison")

    def _kill_compare_process(self):
        """
        Kills the compare process
        :raise exceptions.OSError
        """
        os.kill(self._compare_process.pid, signal.SIGKILL)

    def _create_or_recycle_player_process_if_needed(self):
        """
        Creates a new player resources if one of the following conditions is met:
        (1) This is the first time this method is invoked, hence there is no process already running
        (2) The age of the process (number of playbacks it ran) exceeds the recycle rate
        """
        # Process too old, recycle
        if self._compare_process is not None and \
                self._compare_process_age >= self.compare_execution_config.compare_process_recycle_rate:
            # Signal process to terminate and wait for it to do so
            self._terminate_process.set()
            self._compare_process.join()
            self._compare_process = None
            # Reset terminate state
            self._terminate_process.clear()

        if self._compare_process is None:
            self._create_new_player_process()

        # Increase process age
        self._compare_process_age += 1

    def _create_new_player_process(self):
        """
        Creates and start new player process, ready to take playback tasks
        """
        self._compare_process = mp.Process(
            target=self._playback_process_target, name='Playback runner')
        self._compare_process.start()
        self._compare_process_age = 0

    def _playback_process_target(self):
        """
        Entry point for the playback process (target function),
        """
        while not self._terminate_process.is_set():
            try:
                recording_id = self._compare_tasks.get(True, 0.05)
                try:
                    execution_result = self._play_and_compare_recording(recording_id)
                    self._compare_results.put((True, execution_result))
                except Exception as ex:  # pylint: disable=broad-except
                    logging.info(u'Failure during play and compare in playback process of id {} - {}'.format(
                        recording_id, ex))
                    self._compare_results.put((False, str(ex)))
            except mp.queues.Empty:
                # Every 0.05 second the process will get this as we poll with 0.05 second timeout
                # (in order to listen to termination event)
                pass

    def _play_and_compare_recording(self, recording_id):
        """
        Plays the given recording id and compare its results
        :param recording_id: Recording id to run and compare results
        :type recording_id: str
        :return: Comparison result, playback result, is recorded result an exception,
        is playback result an exception
        :rtype: PlayAndCompareResult
        """
        playback = None
        recorded_result_is_exception = None
        playback_result_is_exception = None

        try:
            playback = self.player(recording_id)

            recorded_result = self.result_extractor(playback.recorded_outputs)
            playback_result = self.result_extractor(playback.playback_outputs)

            recorded_result_is_exception = isinstance(recorded_result, Exception)
            playback_result_is_exception = isinstance(playback_result, Exception)

            comparison_data = {} if self.comparison_data_extractor is None else \
                self.comparison_data_extractor(playback.original_recording)

            comparator_result = self.comparator(recorded_result, playback_result, **comparison_data)
            if not isinstance(comparator_result, ComparatorResult):
                comparator_result = ComparatorResult(comparator_result)

            return PlayAndCompareResult(comparator_result,
                                        playback,
                                        recorded_result_is_exception,
                                        playback_result_is_exception)

        except Exception as ex:  # pylint: disable=broad-except
            return PlayAndCompareResult(ComparatorResult.failure_result(recording_id, ex),
                                        playback,
                                        recorded_result_is_exception,
                                        playback_result_is_exception)

    @staticmethod
    def _comparison_stats_repr(counter):
        """
        :param counter:
        :type counter: collections.Counter
        :return: Representation of comparison statistics
        :rtype: str
        """
        return u'comparison stats: (equal - {}, fixed - {}, diff - {}, failed - {}, equalizer failures - {}'.format(
            counter[EqualityStatus.Equal],
            counter[EqualityStatus.Fixed],
            counter[EqualityStatus.Different],
            counter[EqualityStatus.Failed],
            counter[EqualityStatus.EqualizerFailure])

    @staticmethod
    def get_exception_stack_trace():
        """
        :return: Gets the stack trace as a string of exception raised in current context
        :rtype: str
        """
        __, __, trace_back = sys.exc_info()
        st_file = StringIO()
        traceback.print_tb(trace_back, file=st_file)
        return st_file.getvalue()



import uuid
from collections import OrderedDict
from random import shuffle

from jsonpickle import encode, decode

from playback.recordings.memory.memory_recording import MemoryRecording
from playback.tape_cassette import TapeCassette


class InMemoryTapeCassette(TapeCassette):
    """
    Implementation of TapeCassette that saves everything in memory, mainly for playing around and testing
    and not meant for production use
    """
    def __init__(self):
        self._recordings = OrderedDict()
        self._last_id = None

    def create_new_recording(self, category):
        """
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        return MemoryRecording(u'{}/{}'.format(category, uuid.uuid1().hex))

    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        # We use pickle serialization to find common serialization pitfalls when real cassettes will be used
        self._recordings[recording.id] = encode(recording, unpicklable=True)
        self._last_id = recording.id

    def get_recording(self, recording_id):
        """
        Get recording stored in the given id
        :param recording_id: Id to look for the recording
        :type recording_id: basestring
        :return: Recording in the given id
        :rtype: playback.recording.Recording
        """
        serialized_recording = self._recordings.get(recording_id)
        if serialized_recording is None:
            return None
        deserialized_form = decode(serialized_recording)
        return MemoryRecording(_id=deserialized_form.id, recording_data=deserialized_form.recording_data,
                               recording_metadata=deserialized_form.recording_metadata)

    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                           random_results=False):
        result = []
        for serialized_recording in self._recordings.values():
            recording = decode(serialized_recording)
            if self.extract_recording_category(recording.id) != category:
                continue

            if metadata:
                # Filter based on metadata if provided
                if not TapeCassette.match_against_recorded_metadata(metadata, recording.get_metadata()):
                    continue

            result.append(recording.id)

        if limit:
            result = result[:limit]

        if random_results:
            shuffle(result)

        return iter(result)

    def extract_recording_category(self, recording_id):
        """
        :param recording_id: Recording id to extract category from
        :type recording_id: str
        :return: Recording's category
        :rtype: str
        """
        return recording_id.split('/')[0]

    def get_last_recording_id(self):
        """
        :return: Last recording id
        :rtype: str
        """
        return self._last_id

    def get_all_recording_ids(self):
        """
        :return: All recording ids
        :rtype: list of str
        """
        return sorted(self._recordings.keys())



# pylint: disable=not-context-manager
# pylint: disable=broad-except
# pylint: disable=too-many-lines
from __future__ import absolute_import
from collections import namedtuple, Counter
import logging
from random import Random
from datetime import datetime
from time import time
import threading
from jsonpickle import encode
from decorator import contextmanager

from playback.exceptions import InputInterceptionKeyCreationError, OperationExceptionDuringPlayback, \
    TapeRecorderException, RecordingKeyError
from playback.utils.is_iterable import is_iterable
from playback.utils.pickle_copy import pickle_copy

_logger = logging.getLogger(__name__)


class TapeRecorder(object):
    """
    This class is used to "record" operation and "replay" (rerun) recorded operation on any code version.
    The recording is done by placing different decorators that intercepts the operation, its inputs and outputs by using
    decorators.
    """

    DURATION = '_tape_recorder_recording_duration'
    RECORDED_AT = '_tape_recorder_recorded_at'
    OPERATION_OUTPUT_ALIAS = '_tape_recorder_operation'
    OPERATION_CLASS = '_tape_recorder_operation_class'
    EXCEPTION_IN_OPERATION = '_tape_recorder_exception_in_operation'
    INCOMPLETE_RECORDING = '_tape_recorder_incomplete_recording'

    def __init__(self, tape_cassette, random_seed=None):
        """
        :param tape_cassette: The storage driver to hold the recording in
        :type tape_cassette: playback.tape_cassette.TapeCassette
        :param random_seed: random seed for the sampling rate calculation
        :type random_seed: int
        """
        self.tape_cassette = tape_cassette
        self.recording_enabled = False
        self._active_recording = None
        self._active_recording_parameters = None
        self._playback_recording = None
        self._playback_outputs = []
        self._invoke_counter = Counter()
        self._classes_recording_params = {}
        self._random = Random(random_seed)
        self._force_sample = False
        self._thread_locals = threading.local()

    @contextmanager
    def start_recording(self, category, metadata, post_operation_metadata_extractor=None):
        """
        Starts a recording scope
        :param post_operation_metadata_extractor: Callback used to extract extra metadata once the operation
        is completed and add it to the recording metadata
        :type post_operation_metadata_extractor: function
        :param metadata: Recording metadata
        :type metadata: dict
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        """
        assert self._active_recording is None, 'Cannot start recording while another recording is already running'

        self._active_recording = self.tape_cassette.create_new_recording(category)
        self._active_recording_parameters = self._classes_recording_params.get(
            metadata[TapeRecorder.OPERATION_CLASS], RecordingParameters())
        _logger.info(u'Starting recording for category {} with id {}'.format(category, self._active_recording.id))
        start_time = time()
        try:
            yield
            metadata[TapeRecorder.EXCEPTION_IN_OPERATION] = False
        except Exception:
            metadata[TapeRecorder.EXCEPTION_IN_OPERATION] = True
            raise
        finally:
            # Recording was discarded
            if self._active_recording is not None:
                recording = self._active_recording
                force_sample = self.is_recording_sample_forced
                recording_parameters = self._active_recording_parameters

                # Clear recording not to leave recording in active state if we have
                # some exception raised in following code
                self._reset_active_recording()

                if not self._should_sample_active_recording(recording, recording_parameters, force_sample):
                    self.tape_cassette.abort_recording(recording)
                else:
                    duration = time() - start_time

                    self._add_post_operation_metadata(recording, metadata, post_operation_metadata_extractor, duration)

                    try:
                        self.tape_cassette.save_recording(recording)
                        _logger.info(u'Finished recording of category {} with id {}, recording duration {:.2f}'.format(
                            category, recording.id, duration))
                    except Exception:
                        _logger.exception(u'Failed saving recording of category {} with id {}'.format(
                            category, recording.id))

    def discard_recording(self):
        """
        Discards currently active recording process
        """
        if self._active_recording is not None:
            _logger.info(
                u'Recording with id {} was discarded'.format(self._active_recording.id))
            self.tape_cassette.abort_recording(self._active_recording)
            self._reset_active_recording()

    def force_sample_recording(self):
        """
        Make sure currently active recording will be sampled (unless explicitly discarded or set to ignore enforcement)
        """
        if self._active_recording is not None:
            if self._active_recording_parameters.ignore_enforced_sampling:
                return
            _logger.info(
                u'Recording with id {} sampling is enforced'.format(self._active_recording.id))
            self._force_sample = True

    @property
    def is_recording_sample_forced(self):
        """
        :return: If current recording sampling is enforced, meaning it should be kept regardless of any sampling
        calculation
        :rtype: bool
        """
        return self._force_sample

    @staticmethod
    def _add_post_operation_metadata(recording, metadata, post_operation_metadata_extractor, duration):
        """
        :param recording: Active recording
        :type recording: playback.recording.Recording
        :param metadata: Currently collected metadata
        :type metadata: dict
        :param post_operation_metadata_extractor: Callback used to extract extra metadata once the operation
        is completed and add it to the recording metadata
        :type post_operation_metadata_extractor: function
        :param duration: Recording duration
        :type duration: float
        """
        metadata[TapeRecorder.DURATION] = duration
        metadata[TapeRecorder.RECORDED_AT] = str(datetime.utcnow())
        outputs = TapeRecorder._extract_recorded_output(recording, direct_access=True)
        # Check if the operation has completed by checking if we have operation output recorded, if not it means
        # the operation method didn't complete and regular exception was not caught
        # (this will happen when BaseException is raised such as KeyboardInterrupt, SystemExit)
        incomplete = not any(TapeRecorder.OPERATION_OUTPUT_ALIAS in o.key for o in outputs)
        metadata[TapeRecorder.INCOMPLETE_RECORDING] = incomplete
        if post_operation_metadata_extractor:
            try:
                metadata.update(post_operation_metadata_extractor())
            except Exception:
                _logger.exception(u'Exception caught while extractor post operation metadata for recording id {}, '
                                  u'skipping metadata extraction'.format(recording.id))
        recording.add_metadata(metadata)

    def _reset_active_recording(self):
        """
        Reset recording state, make tape recorder ready for next recording
        """
        self._active_recording = None
        self._active_recording_parameters = None
        self._force_sample = False
        # Clear any previous invocation counter state
        self._invoke_counter = Counter()

    def _record_data(self, key, data):
        """
        Puts current data under given key in the current recording
        :param key: Data key
        :type key: basestring
        :param data: Data to record (it needs to be serializable)
        :type data: Any
        """
        self._assert_recording()
        _logger.info(u'Recording data for recording id {} under key {}'.format(self._active_recording.id, key))
        self._active_recording[key] = data

    def _assert_recording(self):
        """
        Assert there is active recording
        """
        assert self._active_recording is not None, 'No recoding is currently being made'

    def _record_output(self, alias, invocation_number, args, kwargs, data_handler=None):
        """
        Record the given invocation as output
        :param alias: Output alias
        :type alias: str
        :param invocation_number: Current invocation number for the given alias
        :type invocation_number: int
        :param args: Invocation arguments
        :type args: tuple
        :param kwargs: Invocation keyword arguments
        :type kwargs: dict
        :param data_handler: Optional data handler that prepare and restore the output data for and from the recording
        when default pickle serialization is not enough.
        :type data_handler: playback.interception.output_interception.OutputInterceptionDataHandler
        """
        interception_key = self._output_interception_key(alias, invocation_number) + '.output'

        if data_handler:
            try:
                value = data_handler.prepare_output_for_recording(interception_key, args, kwargs)
            except Exception:
                error_message = u'Prepare output for recording error for interception key \'{}\''.format(
                    interception_key)

                _logger.exception(error_message)

                self.discard_recording()
                return
        else:
            value = {'args': list(args), 'kwargs': kwargs}

        if self.in_playback_mode:
            self._playback_outputs.append(Output(interception_key, value))
            return

        # Recording is discarded
        if self._active_recording is None:
            return

        self._record_data(interception_key, value)

    def enable_recording(self):
        """
        Enable recording and interception for all decorated places
        """
        _logger.info('Enabling recording')
        self.recording_enabled = True

    def disable_recording(self):
        """
        Disable recording of all interception in all decorated places
        """
        _logger.info('Disabling recording')
        self.recording_enabled = False

    @property
    def in_recording_mode(self):
        """
        :return: Is in recording mode
        :rtype: bool
        """
        return self.recording_enabled and self._active_recording is not None

    @property
    def in_playback_mode(self):
        """
        :return: Is in playback mode
        :rtype: bool
        """
        return self._playback_recording is not None

    @property
    def current_recording_id(self):
        """
        :return: Returns the id of recording in the current context or None if there is no such recording
        :rtype: basestring or None
        """
        if self.in_recording_mode:
            return self._active_recording.id
        if self.in_playback_mode:
            return self._playback_recording.id
        return None

    @property
    def _currently_in_interception(self):
        """
        :return: Is currently in interception
        :rtype: bool
        """
        if not hasattr(self._thread_locals, 'currently_in_interception'):
            return False

        return self._thread_locals.currently_in_interception

    @_currently_in_interception.setter
    def _currently_in_interception(self, value):
        """
        :param value: Is currently in interception
        :type value: bool
        """
        self._thread_locals.currently_in_interception = value

    @property
    def _should_intercept(self):
        """
        :return: Should run interception
        :rtype: bool
        """
        # We don't want interception inside interception (inception)
        return not self._currently_in_interception and (self.in_recording_mode or self.in_playback_mode)

    def operation(self, metadata_extractor=None):
        """
        :param metadata_extractor: Extracts metadata from the operation when an invocation is recorded
        :type metadata_extractor: function
        :return: Decorated function that should be a recorder operation
        :rtype: function
        """
        return self._operation(class_function=False, metadata_extractor=metadata_extractor)

    def recording_params(self, recording_parameters=None, **kwargs):
        """
        :param recording_parameters: Recording Parameters objects for this operation
        :type sampling_rate: RecordingParameters
        :return: Class decorator that configure specific recording parameter for the operation class
        :rtype: function
        """

        def wrapper(cls):
            self._classes_recording_params[cls] = recording_parameters or RecordingParameters(**kwargs)
            return cls

        return wrapper

    def class_operation(self, metadata_extractor=None):
        """
        :param metadata_extractor: Extracts metadata from the operation when an invocation is recorded
        :type metadata_extractor: function
        :return: Decorated class function that should be a recorder operation
        :rtype: function
        """
        return self._operation(class_function=True, metadata_extractor=metadata_extractor)

    def _operation(self, class_function, metadata_extractor=None):
        """
        :param class_function: Is this a class function or instance function
        :type class_function: bool
        :param metadata_extractor: Extracts metadata at the end of operation when an invocation is
        recorded
        :type metadata_extractor: function
        :return: Decorated function that should be a recorder operation
        :rtype: function
        """

        def func_decoration(func):

            def decorated_function(*args, **kwargs):
                if self.in_playback_mode:
                    return self._execute_operation_func(func, args, kwargs)

                if not self.recording_enabled:
                    return func(*args, **kwargs)

                cls = args[0] if class_function else type(args[0])

                recording_parameters = self._classes_recording_params.get(cls, RecordingParameters())
                if recording_parameters.skipped:
                    return func(*args, **kwargs)

                class_name = cls.__name__

                # As meta add the operation class when possible and class name as category
                metadata = {TapeRecorder.OPERATION_CLASS: cls}

                def post_operation_metadata_extractor():
                    return metadata_extractor(*args, **kwargs)

                post_operation = post_operation_metadata_extractor if metadata_extractor else None

                with self.start_recording(
                        class_name, metadata,
                        post_operation_metadata_extractor=post_operation):
                    return self._execute_operation_func(func, args, kwargs)

            return decorated_function

        return func_decoration

    def _execute_operation_func(self, func, args, kwargs):
        """
        Executes the operation function record its output and return the result
        :param func: Operation function
        :type func: function
        :param args: Execution args
        :type args: tuple
        :param kwargs: Execution kwargs
        :type kwargs: dict
        :return: Function result
        :rtype: Any
        """
        try:
            result = func(*args, **kwargs)
        except TapeRecorderException:
            raise
        except Exception as ex:
            self._record_output(TapeRecorder.OPERATION_OUTPUT_ALIAS, invocation_number=1,
                                args=[self._serializable_exception_form(ex)], kwargs={})

            if self.in_playback_mode:
                # In playback mode we want to capture this as an error that is an output of the function
                # which is a legit recorded result, and not fail the playback it self
                raise OperationExceptionDuringPlayback()

            raise

        # We record the operation result as an output
        self._record_output(TapeRecorder.OPERATION_OUTPUT_ALIAS, invocation_number=1, args=[result],
                            kwargs={})
        return result

    def _should_sample_active_recording(self, recording, recording_parameters, force_sample):
        """
        :param recording: Active recording
        :type recording: playback.recording.Recording
        :param recording_parameters: Current recording parameters
        :type recording_parameters: RecordingParameters
        :param force_sample: Whether current recording sampling is forced

        :type force_sample: bool
        :return: Whether to sample current recording based on its class decoration and sampling calculations
        :rtype: bool
        """
        if force_sample:
            return True

        if recording_parameters.sampling_rate >= 1:
            return True

        sample_value = self._random.random()
        part_of_sample = sample_value <= recording_parameters.sampling_rate
        _logger.info(u'Recording id {} sampled = {} - ({:.3f} <= sampling rate:{:.2f})?'.format(
            recording.id, part_of_sample, sample_value, recording_parameters.sampling_rate))
        return part_of_sample

    @staticmethod
    def _serializable_exception_form(exception):
        """
        :param exception: Exception to return a serializable form
        :type exception: exceptions.Exception
        :return: Serializable form of the exception, if the exception is self is serializable will return the exception
        """
        try:
            encode(exception, unpicklable=True)
            return exception
        except Exception:
            # Upon any failure of encoding, we return a basic form of type of exception and its message
            return {'error_type': type(exception), 'error_repr': repr(exception)}

    def record_data(self, key, value):
        """
        Record given data if in recording mode
        :param key: Data recording key
        :type key: str
        :param value: Value to record (Needs to be serializable)
        :type value: Any
        """
        if not self.in_recording_mode:
            return

        self._record_data(key, value)

    def play_data(self, key):
        """
        Play recorded data if in playback mode
        :param key: Recorded data key
        :type key: str
        :return: Recorded data under given key or None if not in playback mode
        :rtype: Any
        :raise: playback.exceptions.RecordingKeyError
        """
        if not self.in_playback_mode:
            return None

        return self._playback_recording.get_data(key)

    def static_intercept_input(self, alias, alias_params_resolver=None, data_handler=None, capture_args=None,
                               run_intercepted_when_missing=False, value_when_missing=None, fallback_aliases=None):
        """
        Decorates a static function that acts as an input to the operation, the result of the function is
        the  recorded input and the passed arguments and function name (or alias) or used as key for the input
        :param alias: Input alias, used to uniquely identify the input function, hence the name should be unique across
        all relevant inputs this operation can reach. This should be renamed as it will render previous recording
        useless
        :type alias: str
        :param alias_params_resolver: Optional function that resolve parameters inside alias if such are given,
        this is useful when you have the same input method invoked many times with the same arguments on different class
        instances
        :type alias_params_resolver: function
        :param data_handler: Optional data handler that prepare and restore the input data for and from the recording
        when default pickle serialization is not enough
        :type data_handler: playback.interception.input_interception.InputInterceptionDataHandler
        :param capture_args: If a list is given, it will annotate which arg indices and/or names should be
        captured as part of the intercepted key (invocation identification). If None, all args are captured
        :type capture_args: list of CapturedArg
        :param run_intercepted_when_missing: If no matching content is found on recording during playback,
        run the original intercepted method
        :type run_intercepted_when_missing: bool
        :param value_when_missing: If no matching content is found, returns the given value instead. If the given value
        is a function, it will be invoked with the arguments passed to the intercepted method and the invocation value
        will be returned
        :type value_when_missing: function or Any
        :param fallback_aliases: A list of fallback aliases or a function returning such a list. These aliases will be
        used to find the data in the recording in case when there are no results for the main alias. Useful if the main
        alias was changed (for example, due to a refactoring), but old recordings should still be playbable.
        :type fallback_aliases: function or list of str
        :return: Decorated function
        :rtype: function
        """
        return self._intercept_input(alias, alias_params_resolver, data_handler, capture_args,
                                     run_intercepted_when_missing, value_when_missing, fallback_aliases,
                                     static_function=True)

    def intercept_input(self, alias, alias_params_resolver=None, data_handler=None, capture_args=None,
                        run_intercepted_when_missing=False, value_when_missing=None, fallback_aliases=None):
        """
        Decorates a function that that acts as an input to the operation, the result of the function is the
        recorded input and the passed arguments and function name (or alias) or used as key for the input
        :param alias: Input alias, used to uniquely identify the input function, hence the name should be unique across
        all relevant inputs this operation can reach. This should be renamed as it will render previous recording
        useless
        :type alias: str
        :param alias_params_resolver: Optional function that resolve parameters inside alias if such are given,
        this is useful when you have the same input method invoked many times with the same arguments on different class
        instances
        :type alias_params_resolver: function
        :param data_handler: Optional data handler that prepare and restore the input data for and from the recording
        when default pickle serialization is not enough
        :type data_handler: playback.interception.input_interception.InputInterceptionDataHandler
        :param capture_args: If a list is given, it will annotate which arg indices and/or names should be
        captured as part of the intercepted key (invocation identification). If None, all args are captured
        :type capture_args: list of CapturedArg
        :param run_intercepted_when_missing: If no matching content is found on recording during playback,
        run the original intercepted method
        :type run_intercepted_when_missing: bool
        :param value_when_missing: If no matching content is found, returns the given value instead. If the given value
        is a function, it will be invoked with the arguments passed to the intercepted method and the invocation value
        will be returned
        :type value_when_missing: function or Any
        :param fallback_aliases: A list of fallback aliases or a function returning such a list. These aliases will be
        used to find the data in the recording in case when there are no results for the main alias. Useful if the main
        alias was changed (for example, due to a refactoring), but old recordings should still be playbable.
        :type fallback_aliases: function or list of str
        :return: Decorated function
        :rtype: function
        """
        return self._intercept_input(alias, alias_params_resolver, data_handler, capture_args,
                                     run_intercepted_when_missing, value_when_missing, fallback_aliases,
                                     static_function=False)

    def static_intercept_output(self, alias, data_handler=None, fail_on_no_recorded_result=True,
                                default_result_when_not_recorded=None):
        """
        Decorates a static function that that acts as an output of the operation, the arguments are recorded as the
        output and the result of the function is captured
        :param alias: Output alias, used to uniquely identify the input function, hence the name should be unique
        across all relevant inputs this operation can reach. This should be renamed as it will render previous
        recording useless
        :type alias: str
        :param data_handler: Optional data handler that prepare and restore the output data for and from the recording
        when default pickle serialization is not enough.
        :type data_handler: playback.interception.output_interception.OutputInterceptionDataHandler
        :param fail_on_no_recorded_result: Whether to fail if there is no recording of a result or return None.
        Setting this to False is useful when there are already pre existing recording and this is a new output
        interception while we want to be able to playback old recordings and the return value of the output is not
        actually used.
        :type fail_on_no_recorded_result: bool
        :param default_result_when_not_recorded: Which result to return if there is no recording of a result, for this
        to take affect 'fail_on_no_recorded_result' must be set to False
        :type default_result_when_not_recorded: tuple
        :return: Decorated function
        :rtype: function
        """
        return self._intercept_output(alias, data_handler, fail_on_no_recorded_result, default_result_when_not_recorded,
                                      static_function=True)

    def intercept_output(self, alias, data_handler=None, fail_on_no_recorded_result=True,
                         default_result_when_not_recorded=None):
        """
        Decorates a function that that acts as an output of the operation, the arguments are recorded as the output and
        the result of the function is captured
        output and the result of the function is captured
        :param alias: Output alias, used to uniquely identify the input function, hence the name should be unique
        across all relevant inputs this operation can reach. This should be renamed as it will render previous
        recording useless
        :type alias: str
        :param data_handler: Optional data handler that prepare and restore the output data for and from the recording
        when default pickle serialization is not enough.
        :type data_handler: playback.interception.output_interception.OutputInterceptionDataHandler
        :param fail_on_no_recorded_result: Whether to fail if there is no recording of a result or return None.
        Setting this to False is useful when there are already pre existing recordings and this is a new output
        interception while we want to be able to playback old recordings and the return value of the output is not
        actually used. Defaults to True
        :type fail_on_no_recorded_result: bool
        :param default_result_when_not_recorded: Which result to return if there is no recording of a result, for this
        to take affect 'fail_on_no_recorded_result' must be set to False
        :type default_result_when_not_recorded: tuple
        :return: Decorated function
        :rtype: function
        """
        return self._intercept_output(alias, data_handler, fail_on_no_recorded_result, default_result_when_not_recorded,
                                      static_function=False)

    def _intercept_output(self, alias, data_handler, fail_on_no_recorded_result, default_result_when_not_recorded,
                          static_function):
        """
        Decorates a function that that acts as an output of the operation, the arguments are recorded as the output and
        the result of the function is captured
        output and the result of the function is captured
        :param alias: Output alias, used to uniquely identify the input function, hence the name should be unique
        across all relevant inputs this operation can reach. This should be renamed as it will render previous
        recording useless
        :type alias: str
        :param data_handler: Optional data handler that prepare and restore the output data for and from the recording
        when default pickle serialization is not enough.
        :type data_handler: playback.interception.output_interception.OutputInterceptionDataHandler
        :param fail_on_no_recorded_result: Whether to fail if there is no recording of a result or return None.
        Setting this to False is useful when there are already pre existing recording and this is a new output
        interception while we want to be able to playback old recordings and the return value of the output is not
        actually used.
        :type fail_on_no_recorded_result: bool
        :param default_result_when_not_recorded: Which result to return if there is no recording of a result, for this
        to take affect 'fail_on_no_recorded_result' must be set to False
        :type default_result_when_not_recorded: tuple
        :param static_function: Is this a static function
        :type static_function: bool
        :return: Decorated function
        :rtype: function
        """

        def func_decoration(func):

            def decorated_function(*args, **kwargs):
                if not self._should_intercept:
                    return func(*args, **kwargs)

                # If same alias (function) is invoked more than once we want to track each output invocation
                self._invoke_counter[alias] += 1
                invocation_number = self._invoke_counter[alias]

                # Both in recording and playback mode we record what is sent to the output
                self._record_output(alias, invocation_number, args if static_function else args[1:], kwargs,
                                    data_handler)

                # Record output may have failed and discarded current recording which would make should intercept false
                if not self._should_intercept:
                    return func(*args, **kwargs)

                interception_key = self._output_interception_key(alias, invocation_number) + '.result'

                if self.in_playback_mode:
                    # Return recording of input invocation
                    try:
                        return self._playback_recorded_interception([interception_key], args, kwargs)
                    except RecordingKeyError:
                        if fail_on_no_recorded_result:
                            raise
                        return default_result_when_not_recorded

                # Record the output result so it can be returned in playback mode
                return self._execute_func_and_record_interception(func, interception_key, args, kwargs)

            return decorated_function

        return func_decoration

    def _intercept_input(self, alias, alias_params_resolver, data_handler, capture_args, run_intercepted_when_missing,
                         value_when_missing, fallback_aliases, static_function):
        """
        Decorates a function that that acts as an input to the operation, the result of the function is the
        recorded input and the passed arguments and function name (or alias) or used as key for the input
        :param alias: Input alias, used to uniquely identify the input function, hence the name should be unique across
        all relevant inputs this operation can reach. This should be renamed as it will render previous recording
        useless
        :type alias: str
        :param alias_params_resolver: Optional function that resolve parameters inside alias if such are given,
        this is useful when you have the same input method invoked many times with the same arguments on different class
        instances
        this is useful when you have the same input method invoked many times on different class instances
        :type alias_params_resolver: function
        :param data_handler: Optional data handler that prepare and restore the input data for and from the recording
        when default pickle serialization is not enough
        :type data_handler: playback.interception.input_interception.InputInterceptionDataHandler
        :param capture_args: If a list is given, it will annotate which arg indices and/or names should be
        captured as part of the intercepted key (invocation identification). If None, all args are captured
        :type capture_args: list of CapturedArg
        :param run_intercepted_when_missing: If no matching content is found on recording during playback,
        run the original intercepted method
        :type run_intercepted_when_missing: bool
        :param value_when_missing: If no matching content is found, returns the given value instead. If the given value
        is a function, it will be invoked with the arguments passed to the intercepted method and the invocation value
        will be returned
        :type value_when_missing: function or Any
        :param static_function: Is this a static function
        :type static_function: bool
        :return: Decorated function
        :rtype: function
        """

        def func_decoration(func):

            is_property = isinstance(func, property)
            if is_property:
                func = func.__get__

            def decorated_function(*args, **kwargs):
                if not self._should_intercept:
                    return func(*args, **kwargs)

                try:
                    formatted_alias = self._format_alias(alias, alias_params_resolver, *args, **kwargs)
                    interception_key = self._input_interception_key(formatted_alias, capture_args,
                                                                    static_function, *args, **kwargs)

                    if callable(fallback_aliases):
                        fallback_aliases_list = fallback_aliases(*args, **kwargs)
                    elif is_iterable(fallback_aliases):
                        fallback_aliases_list = fallback_aliases
                    else:
                        fallback_aliases_list = []

                    possible_keys = [interception_key] + [self._input_interception_key(fallback_alias, capture_args,
                                                                                       static_function, *args, **kwargs)
                                                          for fallback_alias in fallback_aliases_list]
                except Exception as ex:
                    error_message = u'Input interception key creation error for alias \'{}\' - {}'.format(
                        alias, repr(ex))

                    if self.in_playback_mode:
                        raise InputInterceptionKeyCreationError(error_message.encode('utf-8'))

                    _logger.exception(error_message)

                    interception_key = None
                    self.discard_recording()

                if self.in_playback_mode:
                    # Return recording of input invocation
                    try:
                        return self._playback_recorded_interception(possible_keys, args, kwargs, data_handler)
                    except RecordingKeyError:
                        if run_intercepted_when_missing:
                            # Run the original method when content was missing in recording
                            return func(*args, **kwargs)
                        if value_when_missing:
                            if callable(value_when_missing):
                                return value_when_missing(*args, **kwargs)
                            return value_when_missing
                        raise

                return self._execute_func_and_record_interception(func, interception_key, args, kwargs, data_handler)

            return property(decorated_function) if is_property else decorated_function

        return func_decoration

    @staticmethod
    def _format_alias(alias, alias_params_resolver, *args, **kwargs):
        """
        Formats the alias applying alias name resolver if provided
        :param alias: Alias to format
        :type alias: basestring
        :param alias_params_resolver: Optional function that resolve parameters inside alias if such are given,
        this is useful when you have the same input method invoked many times with the same arguments on different class
        instances
        :type alias_params_resolver: function
        :param args: Invocation args
        :type args: tuple
        :param kwargs: Invocation kwargs
        :type kwargs: dict
        :return: Formatted alias
        :rtype: basestring
        """
        if not alias_params_resolver:
            return alias

        return alias.format(**alias_params_resolver(*args, **kwargs))

    def _playback_recorded_interception(self, possible_keys, args, kwargs, data_handler=None):
        """
        Playback the recorded data (value or exception) under the given possible keys
        :param possible_keys: Possible interception keys of a recorded data
        :type possible_keys: list of basestring
        :param args: invocation args
        :type args: tuple
        :param kwargs: invocation kwrags
        :type kwargs: dict
        :param data_handler: Optional data handler that prepare and restore the input data for and from the recording
        :type data_handler: playback.interception.input_interception.InputInterceptionDataHandler
        :return: Recorded intercepted value
        """
        recording_keys = self._playback_recording.get_all_keys()

        interception_key = next((x for x in possible_keys if x in recording_keys), None)

        if not interception_key:
            raise RecordingKeyError(
                u'None of the possible keys \'{}\' was found in the recording'.format(possible_keys).encode("utf-8"))

        recorded = self._playback_recording.get_data(interception_key)
        if 'exception' in recorded:
            raise recorded['exception']

        value = recorded['value']
        if data_handler:
            value = data_handler.restore_input_from_recording(value, args, kwargs)
        return value

    @contextmanager
    def _enter_interception_context(self):
        """
        Activate currently in interception mode so inner methods that also marked for interception will not be
        intercepted as their output/intput is already captured by the wrapping interception
        """
        assert not self._currently_in_interception
        self._currently_in_interception = True
        try:
            yield
        finally:
            self._currently_in_interception = False

    def _execute_func_and_record_interception(self, func, interception_key, args, kwargs, data_handler=None):
        """
        Executes the given function and record the result/exception of the outcome
        :param func: Function to execute
        :type func: function
        :param interception_key: Key to record the data under
        :type interception_key: basestring
        :param args: invocation args
        :type args: tuple
        :param kwargs: invocation kwrags
        :type kwargs: dict
        :param data_handler: Optional data handler that prepare and restore the input data for and from the recording
        :type data_handler: playback.interception.input_interception.InputInterceptionDataHandler
        :return: Invocation result
        """
        # Mark that this invocation is under interception context so any inner interception will be skipped
        with self._enter_interception_context():
            try:
                result = func(*args, **kwargs)
            except Exception as ex:
                if interception_key is not None:
                    # Record exception marking it as exception so we know to throw on playback
                    self._record_data(interception_key, {'exception': ex})
                raise

        if interception_key is not None:
            try:
                recorded_result = data_handler.prepare_input_for_recording(interception_key, result, args, kwargs) \
                    if data_handler else result
            except Exception:
                error_message = u'Prepare input for recording error for interception key \'{}\''.format(
                    interception_key)

                _logger.exception(error_message)

                self.discard_recording()
                return result

            if self._active_recording_parameters.copy_data_on_intercepion:
                try:
                    recorded_result = pickle_copy(recorded_result)
                except Exception as ex:
                    _logger.warning(u"recorded data couldn't be copied (type={} exception={})".format(
                        type(recorded_result), repr(ex)))

            # Record result
            self._record_data(interception_key, {'value': recorded_result})

        return result

    def play(self, recording_id, playback_function):
        """
        Play again the recorder operation on current code
        :param recording_id: Id of the recording
        :type recording_id: basestring
        :param playback_function: A function that plays back the operation using the recording in the given id
        :type playback_function: function
        :return: Playback result of rerunning the recorded operation
        :rtype: Playback
        """
        recording = self.tape_cassette.get_recording(recording_id)
        self._playback_recording = recording
        start = time()

        try:
            playback_function(recording)
        except OperationExceptionDuringPlayback:
            # This is an exception that was raised by the played back function, should be treated as part of the
            # recording output
            pass
        finally:
            playback_duration = time() - start
            playback_outputs = self._playback_outputs
            self._playback_recording = None
            self._playback_outputs = []
            # Clear any previous invocation counter state
            self._invoke_counter = Counter()

        recorded_duration = recording.get_metadata()[TapeRecorder.DURATION]
        recorded_outputs = self._extract_recorded_output(recording)
        return Playback(playback_outputs, playback_duration, recorded_outputs, recorded_duration, recording)

    @staticmethod
    def _extract_recorded_output(recording, direct_access=False):
        """
        :param recording:
        :type recording: playback.recording.Recording
        :param direct_access: Should the output be extracted using direct access
        :type direct_access: bool
        :return: List of output objects
        :rtype: list of Output
        """
        all_output_keys = [key for key in recording.get_all_keys() if key.startswith('output:') and
                           not key.endswith('result')]
        return [Output(key, (recording.get_data if not direct_access else recording.get_data_direct)(key))
                for key in all_output_keys]

    @staticmethod
    def _input_interception_key(alias, capture_args, static_function, *args, **kwargs):
        """
        Creates a key that uniquely represent this input invocation based on alias and invocation arguments
        :param alias: Input alias, used to uniquely identify the input function, hence the name should be unique across
        all relevant inputs this operation can reach. This should be renamed as it will render previous recording
        useless
        :type alias: str
        :param static_function: Is function static
        :type static_function: bool
        :param args: invocation args
        :type args: list
        :param kwargs: invocation kwargs
        :type kwargs: dict
        :param capture_args: If a list is given, it will annotate which arg indices and/or names should be
        captured as part of the intercepted key (invocation identification). If None, all args are captured
        :type capture_args: list of CapturedArg
        :return: Interception key
        :rtype: basestring
        """
        # Capture all args
        if capture_args is None:
            # Attempt to remove self
            args_for_keys = args[1:] if not static_function else args
            kwargs_for_key = kwargs
        # Set to not capture any args
        elif not capture_args:
            args_for_keys = []
            kwargs_for_key = {}
        else:
            args_for_keys = []
            kwargs_for_key = {}
            for captured_arg in capture_args:
                # First check if argument is in kwargs, these covers both cases where item is only defined is kwarg or
                # that it was passed as a kwarg even though it is a mandatory argument
                if captured_arg.name in kwargs:
                    kwargs_for_key[captured_arg.name] = kwargs[captured_arg.name]
                # If None it means this was captures as kwarg only, otherwise this is a mandatory argument that must
                # have an index
                elif captured_arg.position is not None:
                    args_for_keys.append(args[captured_arg.position])

        args_key = encode(args_for_keys, unpicklable=True)
        kwargs_key = encode(sorted(list(kwargs_for_key.items()), key=lambda k_v: k_v[0]), unpicklable=True)
        return u'input: {} args={}, kwargs={}'.format(alias, args_key, kwargs_key)

    @staticmethod
    def _output_interception_key(alias, invocation_number):
        """
        Creates a key that uniquely represents the output invocation based on alias in invocation count
        :param alias: output alias
        :type alias: str
        :param invocation_number: Current invocation number
        :type invocation_number: int
        :return: Interception key
        :rtype: basestring
        """
        return u'output: {} #{}'.format(alias, invocation_number)


CapturedArg = namedtuple('CapturedArg', 'position name')

Output = namedtuple('Output', 'key value')


class RecordingParameters(object):
    def __init__(self, sampling_rate=1.0, ignore_enforced_sampling=False,
                 skipped=False, copy_data_on_intercepion=False):
        """
        :param sampling_rate: Optional sampling rate (between 0 and 1) to applied on recording. Default is 1
        :type sampling_rate: float
        :param ignore_enforced_sampling: Whether to ignore enforce sample explicit calls
        :type ignore_enforced_sampling: bool
        :param skipped: This class should be skipped for recording
        :type skipped: bool
        :param copy_data_on_intercepion: copy each intercepted
               value while recording to prevent mutations, impacts performance
        :type copy_data_on_intercepion: bool
        """
        self.sampling_rate = sampling_rate
        self.ignore_enforced_sampling = ignore_enforced_sampling
        self.skipped = skipped
        self.copy_data_on_intercepion = copy_data_on_intercepion


class Playback(object):
    def __init__(self, playback_outputs, playback_duration, recorded_outputs, recorded_duration, original_recording):
        """
        :param playback_outputs: Outputs captured during playback
        :type playback_outputs: list of Output
        :param playback_duration: Duration it took to run the playback
        :type playback_duration: float
        :param recorded_outputs: Outputs captured during recording
        :type recorded_outputs: list of Output
        :param recorded_duration: Duration of the original recorded run
        :type recorded_duration: float
        :param original_recording: Recording that was used to run the playback
        :type original_recording: playback.recording.Recording
        """
        self.playback_outputs = playback_outputs
        self.playback_duration = playback_duration
        self.recorded_outputs = recorded_outputs
        self.recorded_duration = recorded_duration
        self.original_recording = original_recording



from flask import Flask
from flask_restplus import Api

from examples.flask.playback_context import init_recording_mode
from examples.flask.web_services import ContentLengthEndpoint, ContentFirstCharsEndpoint

app = Flask(__name__)
api = Api(app)
api.add_resource(ContentLengthEndpoint, '/content_length')
api.add_resource(ContentFirstCharsEndpoint, '/content_first_chars')


if __name__ == '__main__':
    init_recording_mode()
    app.run(host='0.0.0.0', port=5000, debug=True)



import logging
from functools import reduce
from random import shuffle

import boto3


import pytz


class S3BasicFacade(object):
    def __init__(self, bucket, region=None):
        self.bucket = bucket
        self._bucket = boto3.resource('s3').Bucket(bucket)
        self.client = boto3.client('s3', region_name=region)
        logging.getLogger('botocore').setLevel(logging.CRITICAL)
        logging.getLogger('boto3').setLevel(logging.CRITICAL)
        logging.getLogger('urllib3').setLevel(logging.CRITICAL)
        logging.getLogger('s3transfer').setLevel(logging.CRITICAL)
        logging.getLogger('requests').setLevel(logging.CRITICAL)

    def put_string(self, key, string, **kwargs):
        """
        Put a string in S3

        :param key: S3 key
        :type key: str
        :param string: string to store in S3
        :type string: str
        :param kwargs: kwargs
        :type kwargs: dict
        """

        params = dict(
            Bucket=self.bucket,
            Key=key,
            Body=string
        )
        if kwargs:
            params.update(kwargs)

        return self.client.put_object(**params)

    def get_string(self, key):
        """
        Get the string that associated with the given key from the S3 store.

        :param key: S3 key
        :type key: str
        :return: The string from S3
        :rtype: str
        """
        return self.client.get_object(Bucket=self.bucket, Key=key)['Body'].read()

    def iter_keys(self, prefix=None, start_date=None, end_date=None, content_filter=None, limit=None,
                  random_results=False):
        """
        yields the keys that exist in the S3 store.
        :param prefix: if not None, yields only objects with keys that start with the given prefix.
        :type prefix: str
        :param start_date: Optional last modified start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional last modified end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param content_filter: Function that filters object according to their content
        :type content_filter: function
        :param limit: Optional limit on number of keys to fetch
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :rtype: Iterator[str]
        """

        predicates = []

        if start_date or end_date:
            start_date = pytz.utc.localize(start_date) if start_date else None
            end_date = pytz.utc.localize(end_date) if end_date else None

            predicates.append(lambda o: ((start_date is None or start_date <= o.last_modified) and
                                         (end_date is None or o.last_modified <= end_date)))

        if content_filter:
            predicates.append(lambda o: content_filter(o.get()["Body"].read()))

        if random_results:
            s3_objects = list(self._bucket.objects.filter(Prefix=prefix))
            shuffle(s3_objects)
            s3_objects_iter = iter(s3_objects)
        else:
            s3_objects_iter = self._bucket.objects.filter(Prefix=prefix)

        count = 0
        for s3_object in s3_objects_iter:
            if count == limit:
                break
            # pylint: disable=cell-var-from-loop
            is_relevant = reduce(lambda carry, current: carry and current(s3_object), predicates, True)
            if is_relevant:
                count += 1
                yield s3_object.key

    def delete_by_prefix(self, prefix):
        """
        Deletes all the keys that start with the given prefix in the S3 store.

        :type prefix: str
        """
        return self._bucket.objects.filter(Prefix=prefix).delete()



from __future__ import absolute_import

import random
from copy import copy
from random import Random
from zlib import compress, decompress
import logging
import uuid
from datetime import datetime, timedelta
import json
import six
from jsonpickle import encode, decode
from parse import compile  # pylint: disable=redefined-builtin


from playback.exceptions import NoSuchRecording
from playback.tape_cassette import TapeCassette
from playback.recordings.memory.memory_recording import MemoryRecording
from playback.tape_cassettes.s3.s3_basic_facade import S3BasicFacade
from playback.utils.timing_utils import Timed

_logger = logging.getLogger(__name__)


class S3TapeCassette(TapeCassette):

    FULL_KEY = 'tape_recorder_recordings/{key_prefix}full/{id}'
    METADATA_KEY = 'tape_recorder_recordings/{key_prefix}metadata/{id}'
    RECORDING_ID = '{category}/{day}/{id}'
    DAY_FORMAT = '%Y%m%d'

    def __init__(self, bucket, key_prefix='', region=None, transient=False, read_only=True,
                 infrequent_access_kb_threshold=None, sampling_calculator=None):
        """
        :param bucket: Cassette s3 storage bucket
        :type bucket: str
        :param key_prefix: Optional key prefix for recordings
        :type key_prefix: str
        :param region: Optional aws region
        :type region: str
        :param transient: Is this a transient cassette, all recording under given prefix will be deleted when closed
        (only if not read only)
        :type transient: bool
        :param read_only: If True, this cassette can only be used to fetch recordings and not to create new ones,
        any write operations will raise an assertion.
        :type read_only: bool
        :param infrequent_access_kb_threshold: Threshold in KB that above it object will be saved in STANDARD_IA
        (infrequent access storage class), None means never (default)
        :type infrequent_access_kb_threshold: float
        :param sampling_calculator: Optional sampling ratio calculator function, before saving the recording this
        function will be triggered with (category, recording_size, recording),
        and the function should return a number between 0 and 1 which specify its sampling rate
        :type sampling_calculator: function
        """
        _logger.info(u'Creating S3TapeCassette using bucket {}'.format(bucket))
        self.bucket = bucket
        self.key_prefix = (key_prefix + '/') if key_prefix else ''
        self.transient = transient
        self.read_only = read_only
        self.infrequent_access_threshold = \
            infrequent_access_kb_threshold * 1024 if infrequent_access_kb_threshold else None
        self.sampling_calculator = sampling_calculator
        self._random = Random(110613)
        self._metadata_key_parser = compile(self.METADATA_KEY)
        self._recording_id_parser = compile(self.RECORDING_ID)
        self._s3_facade = S3BasicFacade(self.bucket, region=region)

    def get_recording(self, recording_id):
        """
        Get recording stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: playback.recordings.memory.memory_recording.MemoryRecording
        """
        full_key = self.FULL_KEY.format(key_prefix=self.key_prefix, id=recording_id)
        try:
            _logger.info(u'Fetching compressed recording using key {}'.format(full_key))
            compressed_recording = self._s3_facade.get_string(full_key)
            _logger.info(u'Decompressing recording of key {}'.format(full_key))
            serialized_data = decompress(compressed_recording)
        except Exception as ex:
            if 'NoSuchKey' in type(ex).__name__:
                raise NoSuchRecording(recording_id)
            raise
        _logger.info(u'Decoding recording of key {}'.format(full_key))
        full_data = decode(serialized_data)
        # Extract the meta data also in the full recording
        metadata = full_data.pop('_metadata', {})
        _logger.info(u'Returning recording of key {}'.format(full_key))
        return MemoryRecording(recording_id, recording_data=full_data, recording_metadata=metadata)

    def get_recording_metadata(self, recording_id):
        """
        Get recording's metadata stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: dict
        :raises: playback.exceptions.NoSuchRecording
        """
        metadata_key = self.METADATA_KEY.format(key_prefix=self.key_prefix, id=recording_id)
        try:
            _logger.debug(u'Fetching metadata of recording using key {}'.format(metadata_key))
            serialized_data = self._s3_facade.get_string(metadata_key)
        except Exception as ex:
            if 'NoSuchKey' in type(ex).__name__:
                raise NoSuchRecording(recording_id)
            raise
        _logger.debug(u'Decoding metadata of recording of key {}'.format(metadata_key))
        return decode(serialized_data)

    def create_new_recording(self, category):
        """
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recordings.memory.memory_recording.MemoryRecording
        """
        self._assert_not_read_only()

        _id = self.RECORDING_ID.format(
            category=category,
            day=datetime.today().strftime(self.DAY_FORMAT),
            id=uuid.uuid1().hex
        )
        logging.info(u'Creating a new recording with id {}'.format(_id))
        return MemoryRecording(_id)

    def _assert_not_read_only(self):
        """
        Asserts that current cassette is not in read only mode
        """
        assert not self.read_only, 'Cassette is in readonly mode'

    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recordings.memory.memory_recording.MemoryRecording
        """
        self._assert_not_read_only()

        full_data = copy(recording.recording_data)
        # We put meta data also in the full recording
        full_data['_metadata'] = recording.recording_metadata

        with Timed() as timed:
            encoded_full = encode(full_data, unpicklable=True)
        encoding_duration = timed.duration

        with Timed() as timed:
            if six.PY3 and isinstance(encoded_full, str):
                compressed_full = compress(bytes(encoded_full.encode('utf-8')))
            else:
                compressed_full = compress(encoded_full)
        compression_duration = timed.duration

        recording_size = len(compressed_full)

        if not self._should_sample(recording, recording_size):
            logging.info(u'Recording with id {} is not chosen to be sampled and is being discarded'.format(
                recording.id))
            return

        storage_class = self._calculate_storage_class(recording_size)

        full_key = self.FULL_KEY.format(key_prefix=self.key_prefix, id=recording.id)
        metadata_key = self.METADATA_KEY.format(key_prefix=self.key_prefix, id=recording.id)

        _logger.debug(u"Saving recording full data at bucket {} under key {}".format(self.bucket, full_key))
        self._s3_facade.put_string(full_key, compressed_full, StorageClass=storage_class)
        # We break into two keys so we can do faster and cheap filtering based on metadata not requiring to fetch the
        # entire recording data
        _logger.debug(u"Saving recording metadata at bucket {} under key {}".format(self.bucket, metadata_key))
        self._s3_facade.put_string(metadata_key, encode(recording.recording_metadata, unpicklable=True))
        _logger.info(
            u"Recording saved at bucket {} under key {} "
            u"(recording size: {:.1f}KB -compressed-> {:.1f}KB, storage class: {}, "
            u"encoding/compression durations: {:.2f}/{:.2f})".format(
                self.bucket, full_key,
                len(encoded_full) / 1024.0, len(compressed_full) / 1024.0, storage_class,
                encoding_duration, compression_duration))

    def _calculate_storage_class(self, recording_size):
        """
        :param recording_size: Length of compressed recording full data
        :type recording_size: int
        :return: S3 Storage class to be used based on recording size
        :rtype: str
        """
        storage_class = 'STANDARD'
        if self.infrequent_access_threshold and recording_size >= self.infrequent_access_threshold:
            storage_class = 'STANDARD_IA'
        return storage_class

    def _should_sample(self, recording, recording_size):
        """
        Returns whether the given recording should be part of the sample
        :param recording: Recording to determine sample rate for
        :type recording: MemoryRecording
        :param recording_size: Length of compressed recording full data
        :type recording_size: int
        :return:
        :rtype:
        """
        if self.sampling_calculator is None:
            return True

        category = self.extract_recording_category(recording.id)

        ratio = self.sampling_calculator(category, recording_size, recording)
        if ratio >= 1:
            return True

        return self._random.random() <= ratio

    def create_id_prefix_iterators(self, id_prefixes, start_date=None, end_date=None, content_filter=None, limit=None,
                                   random_results=False):
        """
        Creates a list of iterators for every day in case of using dates or for category otherwise.
        :param id_prefixes: list of prefixes to use
        :type id_prefixes: list of basestring
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param content_filter: Optional limit on number of ids to fetch
        :type content_filter: function
        :param limit: Optional filter function to use
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :return: list of Iterator of keys matching the given parameters
        :rtype: list of collections.Iterator[basestring]
        """
        return [self._s3_facade.iter_keys(
            prefix=self.METADATA_KEY.format(
                key_prefix=self.key_prefix, id=id_prefix
            ),
            start_date=start_date,
            end_date=end_date,
            content_filter=content_filter,
            limit=copy(limit),
            random_results=random_results) for id_prefix in id_prefixes]

    @staticmethod
    def _create_content_filter_func(metadata):
        """
        Create a filter function which filters on the metadata values
        :param metadata: metadata values to filter by
        :type metadata: dict
        :return: the filter function
        :rtype: function
        """
        def content_filter_func(recording_str):
            recording_metadata = json.loads(recording_str)
            return TapeCassette.match_against_recorded_metadata(metadata, recording_metadata)

        return content_filter_func

    def _get_id_prefixes(self, category, start_date=None, end_date=None):
        """
        Get recording IDs prefixes
        :param category: Recordings category
        :type category: basestring
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :return: list of id prefixes
        :rtype: list
        """
        # to improve performance when looking for recordings in s3, the date is added to the folder
        # and when a start date is given we can look for specific folders until today (or end_time)
        if start_date:
            end_date = end_date or datetime.utcnow()
            days = [(start_date + timedelta(days=i)) for i in range((end_date - start_date).days + 1)]
            id_prefixes = ['{}/{}/'.format(category, day.strftime(self.DAY_FORMAT)) for day in days]
        else:
            id_prefixes = ['{}/'.format(category)]

        return id_prefixes

    def _get_days_iterators(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                            random_results=False):
        """
        Get days iterators with recording IDs of each day
        :param category: Recordings category
        :type category: basestring
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :return: List of days iterators
        :rtype: list
        """
        content_filter = self._create_content_filter_func(metadata) if metadata else None

        id_prefixes = self._get_id_prefixes(category, start_date, end_date)

        return self.create_id_prefix_iterators(id_prefixes, start_date, end_date, content_filter, limit, random_results)

    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                           random_results=False):
        """
        Creates an iterator of keys matching the given parameters
        :param category: Recordings category
        :type category: basestring
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :return: Iterator of keys matching the given parameters
        :rtype: collections.Iterator[basestring]
        """

        days_iterators = self._get_days_iterators(category, start_date, end_date, metadata, limit, random_results)

        count = 0
        iter_index = 0
        while count != limit and days_iterators:
            if random_results:
                random_day_iterator = random.choice(days_iterators)
            else:
                random_day_iterator = days_iterators[iter_index % len(days_iterators)]
                iter_index += 1
            key = next(random_day_iterator, None)
            if key:
                result = self._metadata_key_parser.parse(key)
                recording_id = result.named['id']
                _logger.info(u'Found filtered recording id {}'.format(recording_id))
                yield recording_id
                count += 1
            else:
                days_iterators.remove(random_day_iterator)

    def extract_recording_category(self, recording_id):
        """
        :param recording_id: Recording id to extract category from
        :type recording_id: str
        :return: Recording's category
        :rtype: str
        """
        result = self._recording_id_parser.parse(recording_id)
        assert result is not None, 'Unrecognized key used'
        return result.named['category']

    def close(self):
        """
        Close this cassette and release any underlying resources, if set to be transient it will delete all recordings
        """
        if self.read_only or not self.transient:
            return

        full_key = self.FULL_KEY.format(key_prefix=self.key_prefix, id='')
        metadata_key = self.METADATA_KEY.format(key_prefix=self.key_prefix, id='')
        _logger.info(u'Deleting all full recordings at bucket {} with prefix {}'.format(self.bucket, full_key))
        self._s3_facade.delete_by_prefix(full_key)
        _logger.info(u'Deleting all metadata recordings at bucket {} with prefix {}'.format(self.bucket, metadata_key))
        self._s3_facade.delete_by_prefix(metadata_key)



def is_iterable(thing):
    """
    Checks if the parameter is iterable.
    :param thing:
    :return:
    """
    try:
        iter(thing)
    except TypeError:
        return False

    return True



import logging
from threading import Event, Lock, Thread
from playback.recordings.memory.memory_recording import MemoryRecording
from playback.tape_cassette import TapeCassette

_logger = logging.getLogger(__name__)


class AsyncRecordOnlyTapeCassette(TapeCassette):
    # pylint: disable=too-many-instance-attributes
    """
    Wraps TapeCassette with asynchronous execution of the operation that change the state of the recording, this
    cassette can only be used for recording and for playback
    """
    def __init__(self, tape_cassette, flush_interval=0.1, timeout_on_close=10):
        """
        :param tape_cassette: The storage driver to hold the recording in and wrap with asynchronous behaviour
        :type tape_cassette: playback.tape_cassette.TapeCassette
        :param flush_interval: Interval to flush recording to the underlying storage (tape cassette)
        :type flush_interval: float
        :param timeout_on_close: How much time to wait for joining recording thread on close
        :type timeout_on_close: float
        """
        self.wrapped_tape_cassette = tape_cassette
        self._flush_interval = flush_interval
        self._timeout_on_close = timeout_on_close
        self._recording_operation_buffer = []
        self._stop_event = Event()
        self._lock = Lock()
        self._update_recording_thread = Thread(target=self._recording_loop, name="AsyncTapeCassette Thread")
        self._update_recording_thread.setDaemon(True)
        self._started = False

    def start(self):
        """
        Starts the recording thread
        """
        _logger.info("Starting AsyncTapeCassette")
        self._started = True
        self._update_recording_thread.start()

    def close(self):
        """
        Signal the cassette to close, this will signal the underlying thread to stop waiting for more recording and
        will join it until its completed sending remaining recordings using the given timeout
        """
        _logger.info("Shutting down AsyncTapeCassette (joining for {}s)".format(self._timeout_on_close))
        self._started = False
        self._stop_event.set()
        try:
            self._update_recording_thread.join(self._timeout_on_close)
        except RuntimeError:
            # If thread was not started
            pass
        self.wrapped_tape_cassette.close()
        _logger.info("AsyncTapeCassette has shutdown")

    def get_recording(self, recording_id):
        raise TypeError("AsyncTapeCassette should only be used for recording, not playback")

    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                           random_results=False):
        raise TypeError("AsyncTapeCassette should only be used for recording, not playback")

    def extract_recording_category(self, recording_id):
        raise TypeError("AsyncTapeCassette should only be used for recording, not playback")

    def create_new_recording(self, category):
        """
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        assert self._started, "Recording thread is not running"
        # The assumption is that create new recording is not a long running task and hence we can do it synchronously,
        # if that will not be the case the creation it self needs to become async as well
        return AsyncRecording(self.wrapped_tape_cassette.create_new_recording(category), self._add_async_operation)

    def _add_async_operation(self, func):
        """
        Adds operation to be executed asynchronously
        :param func: Operation to execute
        :type func: function
        """
        with self._lock:
            self._recording_operation_buffer.append(func)

    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: AsyncRecording
        """
        self._add_async_operation(lambda: self.wrapped_tape_cassette.save_recording(recording.wrapped_recording))

    def _recording_loop(self):
        """
        Runs the recording loop in a dedicated thread
        """
        _logger.info('Async recording thread started')
        while not self._stop_event.is_set():
            self._flush_recording()
            self._stop_event.wait(self._flush_interval)
        _logger.info('Async recording thread signaled to stop thread, flushing any pending recording')

        # Flush any pending recording
        self._flush_recording()
        _logger.info('Async recording thread stopped')

    def _flush_recording(self):
        """
        Flush current pending recording to the underlying storage, this method is blocking till recording is done
        """
        # We don't want to keep the lock while doing actual recording, hence we copy the buffer state and release
        # the lock
        _logger.debug('Flushing pending recording operations')
        with self._lock:
            current_flushed_operations = self._recording_operation_buffer
            self._recording_operation_buffer = []

        # Execute actual recording
        for recording_operation in current_flushed_operations:
            try:
                recording_operation()
            except Exception as ex:  # pylint: disable=broad-except
                _logger.exception(u"Error running recording operation - {}".format(ex))


class AsyncRecording(MemoryRecording):
    """
    Wraps a recording with asynchronous set behaviour, keep recording data transient in memory during the recording for
    being able to fetch data from the recording or the metadata if needed
    """

    def __init__(self, wrapped_recording, add_async_operation_callback):
        """
        :param wrapped_recording: Recording to wrap with asynchronous set data
        :type wrapped_recording: Recording
        :param add_async_operation_callback: A callback to add operations to be executed asynchronously
        :type add_async_operation_callback: function
        """
        # This cassette is only used for recording, hence it has no use of keeping the playback factory
        super(AsyncRecording, self).__init__(wrapped_recording.id)
        self.wrapped_recording = wrapped_recording
        self._add_async_operation_callback = add_async_operation_callback

    def _set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        super(AsyncRecording, self)._set_data(key, value)
        self._add_async_operation_callback(lambda: self.wrapped_recording.set_data(key, value))

    def _add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        super(AsyncRecording, self)._add_metadata(metadata)
        self._add_async_operation_callback(lambda: self.wrapped_recording.add_metadata(metadata))



import random
from datetime import datetime, timedelta
from uuid import uuid1

from playback.studio.equalizer import Equalizer, ComparatorResult, EqualityStatus
from playback.studio.recordings_lookup import RecordingLookupProperties, find_matching_recording_ids
from playback.tape_cassettes.in_memory.in_memory_tape_cassette import InMemoryTapeCassette
from playback.tape_recorder import TapeRecorder

tape_cassette = InMemoryTapeCassette()
tape_recorder = TapeRecorder(tape_cassette)


class ServiceOperation(object):

    @tape_recorder.operation()
    def execute(self):
        """
        Executes the operation and return the key of where the result is stored
        """
        data = self.get_request_data()
        result = self.do_something_with_input(data)
        storage_key = self.store_result(result)
        return storage_key

    @tape_recorder.intercept_input(alias='service_operation.get_request_data')
    def get_request_data(self):
        """
        Reads the required input for the operation
        """
        # Fake input that will be captured in the recording
        return random.randint(0, 10)

    @tape_recorder.intercept_output(alias='service_operation.store_result')
    def store_result(self, result):
        """
        Stores the operation result and return the key that can be used to fetch the result
        """
        result_key = self.put_result_in_mongo(result)
        return result_key

    def do_something_with_input(self, input):
        """
        Apply some logic on input
        """
        return input * 2

    def put_result_in_mongo(self, result):
        """
        Stores the result in mongo
        """
        # Fake output, we don't really need to store anything in the example, return a fake document id
        return uuid1().hex


# Run the operation 5 times and record it
tape_recorder.enable_recording()

for __ in range(10):
    ServiceOperation().execute()


def playback_function(recording):
    """
    Given a recording, replay the recorded operation
    """
    operation_class = recording.get_metadata()[TapeRecorder.OPERATION_CLASS]
    return operation_class().execute()


# Replay last recorded operation
tape_recorder.play(tape_cassette.get_last_recording_id(), playback_function)


# Creates an iterator over relevant recordings which are ready to be played
lookup_properties = RecordingLookupProperties(start_date=datetime.utcnow() - timedelta(days=7),
                                              limit=5)
recording_ids = find_matching_recording_ids(tape_recorder, ServiceOperation.__name__, lookup_properties)


def result_extractor(outputs):
    """
    Given recording or playback outputs, find the relevant output which is the result that needs to be compared
    """
    # Find the relevant captured output
    output = next(o for o in outputs if 'service_operation.store_result' in o.key)
    # Return the captured first arg as the result that needs to be compared
    return output.value['args'][0]


def comparator(recorded_result, replay_result):
    """
    Compare the operation captured output result
    """
    if recorded_result == replay_result:
        return ComparatorResult(EqualityStatus.Equal, "Value is {}".format(recorded_result))
    return ComparatorResult(EqualityStatus.Different,
                            "{recorded_result} != {replay_result}".format(
                                recorded_result=recorded_result, replay_result=replay_result))


def player(recording_id):
    return tape_recorder.play(recording_id, playback_function)


# Run comparison and output comparison result using the Equalizer
equalizer = Equalizer(recording_ids, player, result_extractor, comparator)

for comparison_result in equalizer.run_comparison():
    print('Comparison result {recording_id} is: {result}'.format(
        recording_id=comparison_result.playback.original_recording.id,
        result=comparison_result.comparator_status))



import os
import uuid
import io

import six
from jsonpickle import encode, decode

from playback.exceptions import NoSuchRecording
from playback.recordings.memory.memory_recording import MemoryRecording
from playback.tape_cassette import TapeCassette


class FileBasedTapeCassette(TapeCassette):
    """
    Implementation of TapeCassette that saves each recording in a file, mainly for playing around and testing,
    this is not production grade implementation.
    """
    def __init__(self, directory):
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self.directory = directory

    def get_recording(self, recording_id):
        """
        Get recording stored in the given id
        :param recording_id: Id to look for the recording
        :type recording_id: basestring
        :return: Recording in the given id
        :rtype: playback.recording.Recording
        """
        file_path = self._get_recording_file_path(recording_id)
        if not os.path.isfile(file_path):
            raise NoSuchRecording(recording_id)

        with io.open(file_path, "r", encoding="utf-8") as f:
            encoded = f.read()
        deserialized_form = decode(encoded)
        return MemoryRecording(_id=deserialized_form.id, recording_data=deserialized_form.recording_data,
                               recording_metadata=deserialized_form.recording_metadata)

    def create_new_recording(self, category):
        """
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        return MemoryRecording(u'{}/{}'.format(category, uuid.uuid1().hex))

    def extract_recording_category(self, recording_id):
        """
        :param recording_id: Recording id to extract category from
        :type recording_id: str
        :return: Recording's category
        :rtype: str
        """
        return recording_id.split('/')[0]

    def _save_recording(self, recording):
        """
        Saves the given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        encoded = six.text_type(encode(recording, unpicklable=True))
        with io.open(self._get_recording_file_path(recording.id), "w", encoding="utf-8") as f:
            f.write(encoded)

    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                           random_results=False):
        """
        Creates an iterator of recording ids matching the given search parameters
        :param category: Recordings category
        :type category: str
        :param start_date: Not supported and will be ignored
        :type start_date: datetime.datetime
        :param end_date: Not supported and will be ignored
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :return: Iterator of recording ids matching the given parameters
        :rtype: collections.Iterator[str]
        """
        ids = []
        for file_name in os.listdir(self.directory):
            if not file_name.startswith(category):
                continue

            recording_id = file_name.split('.')[0]
            recording = self.get_recording(recording_id)

            if metadata:
                # Filter based on metadata if provided
                if not all(metadata[key] == recording.get_metadata()[key] for key in metadata.keys()):
                    continue

            ids.append(recording.id)

        if limit:
            ids = ids[:limit]

        return iter(ids)

    def _get_recording_file_path(self, recording_id):
        """
        :param recording_id: Recording id
        :type recording_id: str
        :return: File path for given recording id
        :rtype: str
        """
        return os.path.join(self.directory, recording_id.replace('/', '_')) + '.json'



from playback.interception.files.file_interception import FileInterception
from playback.interception.output_interception import OutputInterceptionDataHandler


class OutputInterceptionFileDataHandler(OutputInterceptionDataHandler, FileInterception):
    """
    Intercept file arguments for playback
    """
    def prepare_output_for_recording(self, interception_key, args, kwargs):
        """
        Reads the intercepted file, encode its content as string and return it ready for recording
        :param interception_key: Input interception key
        :type interception_key: basestring
        :param args: Input invocation args
        :type args: tuple
        :param kwargs: Input invocation kwargs
        :type kwargs: dict
        :return: Input result in a form that should be saved in the recording
        :rtype: dict[str, str]
        """
        return self._intercept_file(args, kwargs)

    def restore_output_from_recording(self, recorded_data):
        """
        Create a file from the recorded content and place it in the given file path
        :param recorded_data: Recorded data provided by the prepare method
        :type recorded_data: Any
        :return: A file holder that holds the content that can be extracted to file
        :rtype: InterceptedOutputFileHolder
        """
        file_path, file_content = self._deserialize_file(recorded_data)
        return InterceptedOutputFileHolder(file_content, file_path)


class InterceptedOutputFileHolder(object):
    """
    Holds intercepted output file content
    """
    def __init__(self, file_content, output_file_path):
        """
        :param file_content: File content as string
        :type file_content: basestring
        :param output_file_path: The path of the file when it was output
        :type output_file_path: basestring
        """
        self.file_content = file_content
        self.output_file_path = output_file_path

    def to_file(self, file_path):
        """
        Writes the output file content to the given file path
        :param file_path: Path to write the content to
        :type file_path: basestring
        """
        with open(file_path, "wb") as binary_file:
            binary_file.write(self.file_content)



# p3ready
from __future__ import absolute_import

import random
import unittest
import uuid
from datetime import datetime, timedelta
from mock import patch
import boto3
from mock.mock import Mock
from moto import mock_s3
from playback.tape_recorder import TapeRecorder

from playback.studio.recordings_lookup import find_matching_recording_ids, RecordingLookupProperties

from playback.exceptions import NoSuchRecording
from playback.recording import Recording
import six
from playback.tape_cassettes.s3.s3_tape_cassette import S3TapeCassette
from six.moves import range
TEST_BUCKET = 'test_bucket'


def assert_items_equal(testcase, seq1, seq2, msg=""):
    if six.PY3:
        testcase.assertCountEqual(seq1, seq2, msg)
    else:
        testcase.assertItemsEqual(seq1, seq2, msg)


@mock_s3
class TestS3TapeCassette(unittest.TestCase):

    def setUp(self):
        conn = boto3.resource('s3', region_name='us-east-1')
        # We need to create the bucket since this is all in Moto's 'virtual' AWS account
        conn.create_bucket(Bucket=TEST_BUCKET)
        self.cassette = S3TapeCassette(TEST_BUCKET, key_prefix='tests_' + uuid.uuid1().hex, transient=True,
                                       read_only=False)
        self.tape_recorder = TapeRecorder(self.cassette)
        self.tape_recorder.enable_recording()

    def tearDown(self):
        self.cassette.close()

    def test_create_save_and_fetch_empty_recording(self):
        recording = self.cassette.create_new_recording('test_operation')
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

    def test_create_save_and_fetch_recording_with_data(self):
        recording = self.cassette.create_new_recording('test_operation')
        recording.set_data('key1', 5)
        recording.set_data('key2', {'obj_key1': 2, 'obj_key2': "bla"})
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

        self.assertEqual(5, recording.get_data('key1'))
        self.assertEqual(5, recording['key1'])
        self.assertEqual(recording.get_data('key1'), fetched_recording.get_data('key1'))

        self.assertEqual({'obj_key1': 2, 'obj_key2': "bla"}, recording.get_data('key2'))
        self.assertEqual(recording.get_data('key2'), fetched_recording.get_data('key2'))

        assert_items_equal(self, ['key1', 'key2'], recording.get_all_keys())
        assert_items_equal(self, recording.get_all_keys(), fetched_recording.get_all_keys())

    def test_get_recording_and_get_recording_metadata_non_existing(self):
        with self.assertRaises(NoSuchRecording):
            self.cassette.get_recording('non existing id')

        with self.assertRaises(NoSuchRecording):
            self.cassette.get_recording_metadata('non existing id')

    def test_get_recording_and_get_recording_metadata_other_exception(self):

        def side_affect(*args, **kwargs):
            raise ValueError('some error')

        with patch('playback.tape_cassettes.s3.s3_basic_facade.S3BasicFacade.get_string',
                   side_affect):
            with self.assertRaises(ValueError) as cm:
                self.cassette.get_recording('some id')
            self.assertEqual('some error', str(cm.exception))

            with self.assertRaises(ValueError) as cm:
                self.cassette.get_recording_metadata('some id')
            self.assertEqual('some error', str(cm.exception))

    def test_create_save_and_fetch_recording_with_metadata(self):
        recording = self.cassette.create_new_recording('test_operation')
        metadata = {'key1': 5, 'key2': {'obj_key1': 2, 'obj_key2': "bla"}}
        recording.add_metadata(metadata)
        self.cassette.save_recording(recording)
        fetched_recording = self.cassette.get_recording(recording.id)
        self.assertEqual(recording.id, fetched_recording.id)

        self.assertEqual(metadata, recording.get_metadata())
        self.assertEqual(recording.get_metadata(), fetched_recording.get_metadata())

        fetched_recording_metadata = self.cassette.get_recording_metadata(recording.id)
        self.assertEqual(metadata, fetched_recording_metadata)

    def test_close_transient_true(self):
        prefix = 'tests_' + uuid.uuid1().hex
        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False) as new_cassette:
            recording = new_cassette.create_new_recording('test_operation')
            new_cassette.save_recording(recording)

        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False) as new_cassette:
            with self.assertRaises(NoSuchRecording):
                new_cassette.get_recording(recording.id)

    def test_close_transient_false(self):
        prefix = 'tests_' + uuid.uuid1().hex
        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=False, read_only=False) as new_cassette:
            recording = new_cassette.create_new_recording('test_operation')
            new_cassette.save_recording(recording)

        # Closing this will clean up the garbage
        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False) as new_cassette:
            self.assertIsNotNone(new_cassette.get_recording(recording.id))

    def test_read_only_cassette(self):
        prefix = 'tests_' + uuid.uuid1().hex
        try:
            with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=False, read_only=False) as new_cassette:
                recording = new_cassette.create_new_recording('test_operation')
                new_cassette.save_recording(recording)

            new_cassette = S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=True)
            fetched_recording = new_cassette.get_recording(recording.id)
            self.assertIsNotNone(fetched_recording)
            with self.assertRaises(AssertionError):
                new_cassette.create_new_recording("some_category")
            with self.assertRaises(AssertionError):
                new_cassette.save_recording(fetched_recording)
            new_cassette.close()

            # Check the close didn't delete the previous recording
            new_cassette = S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=True)
            self.assertIsNotNone(new_cassette.get_recording(recording.id))
            new_cassette.close()

        finally:
            # Clean up the garbage
            S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=False, read_only=False).close()

    def test_fetch_recording_ids_by_category(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1')))
        assert_items_equal(self, [recording3.id],
                           list(self.cassette.iter_recording_ids(category='test_operation2')))

    def test_fetch_recording_ids_by_category_and_date(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 start_date=datetime.utcnow() - timedelta(hours=1))))
        assert_items_equal(self, [],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 start_date=datetime.utcnow() + timedelta(hours=1))))

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 end_date=datetime.utcnow() + timedelta(hours=1))))
        assert_items_equal(self, [],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 end_date=datetime.utcnow() - timedelta(hours=1))))

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 start_date=datetime.utcnow() - timedelta(hours=1),
                                                                 end_date=datetime.utcnow() + timedelta(hours=1))))

    def test_fetch_recording_ids_by_category_date_and_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': True})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': False})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 start_date=datetime.utcnow() - timedelta(hours=1),
                                                                 end_date=datetime.utcnow() + timedelta(hours=1),
                                                                 metadata={'property': False})))

    def test_fetch_recordings_metadata_by_category_date_and_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': True})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': False})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [{'property': False}],
                           list(self.cassette.iter_recordings_metadata(
                               category='test_operation1',
                               start_date=datetime.utcnow() - timedelta(hours=1),
                               end_date=datetime.utcnow() + timedelta(hours=1),
                               metadata={'property': False})))

    def test_fetch_recording_ids_with_wildcard_matching_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': 'val11'})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': 'val21'})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': 'val2*'})))

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': 'val*'})))

    def test_fetch_recordings_metadata_with_wildcard_matching_metadata(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': 'val11'})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': 'val21'})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation2')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [{'property': 'val21'}],
                           list(self.cassette.iter_recordings_metadata(category='test_operation1',
                                                                       metadata={'property': 'val2*'})))

        assert_items_equal(self, [{'property': 'val11'}, {'property': 'val21'}],
                           list(self.cassette.iter_recordings_metadata(category='test_operation1',
                                                                       metadata={'property': 'val*'})))

    def test_fetch_recording_ids_with_wildcard_matching_metadata_value_not_set(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': 'val11'})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': 'val21'})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording3)

        assert_items_equal(self, [recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': 'val2*'})))

        assert_items_equal(self, [recording1.id, recording2.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': 'val*'})))

    def test_fetch_recording_ids_with_matching_metadata_value_not_set(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        recording1.add_metadata({'property': 'val11'})
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        recording2.add_metadata({'property': 'val21'})
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation1')
        recording3.add_metadata({'property': 'test'})
        self.cassette.save_recording(recording3)
        recording4 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording4)

        assert_items_equal(self, [recording2.id, recording3.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': ['val2*', 'test']})))

        assert_items_equal(self, [recording1.id, recording2.id, recording3.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': ['val*', 't*']})))
        assert_items_equal(self, [],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': ['try1', 'try2']})))

        assert_items_equal(self, [recording2.id, recording3.id, recording4.id],
                           list(self.cassette.iter_recording_ids(category='test_operation1',
                                                                 metadata={'property': ['val2*', 'test', None]})))

    def test_fetch_recording_ids_by_category_and_limit(self):
        recording1 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        recording3 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording3)

        self.assertEqual(2, len(list(self.cassette.iter_recording_ids(category='test_operation1', limit=2))))

    def test_fetch_recording_ids_by_limit_overall(self):
        yesterday = Mock(wraps=datetime.today)
        yesterday.return_value = datetime.today()-timedelta(days=1)
        with patch('datetime.today', yesterday, create=True):
            recording1 = self.cassette.create_new_recording('test_operation1')
            self.cassette.save_recording(recording1)
        recording2 = self.cassette.create_new_recording('test_operation1')
        self.cassette.save_recording(recording2)
        self.assertEqual(1, len(list(self.cassette.iter_recording_ids(
            category='test_operation1', limit=1, start_date=datetime.utcnow() - timedelta(days=7)))))

    def test_big_recording_storage_type(self):
        prefix = 'tests_' + uuid.uuid1().hex
        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False,
                            infrequent_access_kb_threshold=1) as new_cassette:
            with patch.object(new_cassette._s3_facade, 'put_string', wraps=new_cassette._s3_facade.put_string) \
                    as patched:
                recording = new_cassette.create_new_recording('test_operation')
                recording.set_data('some_data', list(range(10000)))
                new_cassette.save_recording(recording)
                args, kwargs = patched.call_args_list[0]
                self.assertEqual('STANDARD_IA', kwargs['StorageClass'])

        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False,
                            infrequent_access_kb_threshold=1) as new_cassette:
            with patch.object(new_cassette._s3_facade, 'put_string',
                              wraps=new_cassette._s3_facade.put_string) \
                    as patched:
                recording = new_cassette.create_new_recording('test_operation')
                recording.set_data('some_data', list(range(10)))
                new_cassette.save_recording(recording)
                args, kwargs = patched.call_args_list[0]
                self.assertEqual('STANDARD', kwargs['StorageClass'])

    def test_save_with_sampling_ratio(self):
        prefix = 'tests_' + uuid.uuid1().hex

        def sampling_calculator(category, size, r):
            self.assertEqual(category, 'test_operation')
            self.assertIsInstance(r, Recording)
            return 1 if size < 500 else 0

        with S3TapeCassette(TEST_BUCKET, key_prefix=prefix, transient=True, read_only=False,
                            sampling_calculator=sampling_calculator) as new_cassette:
            recording = new_cassette.create_new_recording('test_operation')
            new_cassette.save_recording(recording)
            self.assertIsNotNone(new_cassette.get_recording(recording.id))

            recording = new_cassette.create_new_recording('test_operation')
            recording.set_data('key', list(range(1000)))
            new_cassette.save_recording(recording)
            with self.assertRaises(NoSuchRecording):
                new_cassette.get_recording(recording.id)

    def test_without_random_sample(self):

        class Operation(object):
            def __init__(self, value=None):
                self._value = value

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input

        for i in range(10):
            Operation(i).execute()

        def playback_function(recording):
            return Operation().execute()

        random.seed(12)
        first_list = None
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=False,
                limit=3)
        )
        first_list = list(playable_recordings)
        self.assertEqual(len(first_list), 3)

        random.seed(4)
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=False,
                limit=3)
        )
        second_list = list(playable_recordings)
        self.assertEqual(second_list, first_list)

    def test_random_sample(self):

        class Operation(object):
            def __init__(self, value=None):
                self._value = value

            @property
            @self.tape_recorder.intercept_input('input')
            def input(self):
                return self._value

            @self.tape_recorder.operation()
            def execute(self):
                return self.input

        for i in range(10):
            Operation(i).execute()

        def playback_function(recording):
            return Operation().execute()

        random.seed(12)
        first_list = None
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=True,
                limit=3)
        )
        first_list = list(playable_recordings)
        self.assertEqual(len(first_list), 3)

        random.seed(4)
        start_date = datetime.utcnow() - timedelta(hours=1)
        playable_recordings = find_matching_recording_ids(
            self.tape_recorder,
            category=Operation.__name__,
            lookup_properties=RecordingLookupProperties(
                start_date=start_date,
                random_sample=True,
                limit=3)
        )
        second_list = list(playable_recordings)
        self.assertNotEqual(second_list, first_list)



import unittest
from random import randint
from tempfile import gettempdir
from datetime import datetime
import os
from contextlib import contextmanager

from playback.interception.files.file_interception import FileInterception
from playback.interception.files.input_file_interception import InputInterceptionFileDataHandler
from playback.interception.files.output_file_interception import OutputInterceptionFileDataHandler


def write_to_file(file_name, content):
    """
    Writes text to a file
    """
    with open(file_name, "w") as text_file:
        text_file.write(str(content))


def read_from_file(file_name):
    """
    Reads text from a file
    """
    with open(file_name, "rb") as text_file:
        return text_file.read()


@contextmanager
def temp_file():
    file_name = os.path.join(gettempdir(), '{}_{}tmp'.format(randint(0, 10000), datetime.now()))
    try:
        yield file_name
    finally:
        try:
            os.remove(file_name)
        except OSError:
            pass


@contextmanager
def temp_env(var_name, new_value):
    """
    Sets temporary and restores a given environment variable. The new value would be converted to string by str
    :param var_name: The name on the variable
    :type var_name: basestring
    :param new_value: The new value
    :type new_value: object # the would be converted using str(o)
    """
    variable_defined = var_name in os.environ
    if variable_defined:
        prev_value = os.getenv(var_name)
    os.environ[var_name] = str(new_value)
    try:
        yield
    finally:
        if variable_defined:
            os.environ[var_name] = prev_value
        else:
            del (os.environ[var_name])


class TestInputInterceptionFileDataHandler(unittest.TestCase):

    def test_prepare_and_restore(self):
        data_handler = InputInterceptionFileDataHandler(file_path_arg_index=0, file_path_arg_name='file_name')
        with temp_file() as file_name:
            write_to_file(file_name, 'some random meaningless content')
            prepared_input = data_handler.prepare_input_for_recording('key', '', [file_name], {})
            self.assertEqual(file_name, prepared_input['file_path'])

            with temp_file() as restore_file:
                data_handler.restore_input_from_recording(prepared_input, [restore_file], {})
                self.assertEqual(read_from_file(restore_file), read_from_file(file_name))

        with temp_file() as file_name:
            write_to_file(file_name, 'some random meaningless content')
            prepared_input = data_handler.prepare_input_for_recording('key', '', [], {'file_name': file_name})
            self.assertEqual(file_name, prepared_input['file_path'])

            with temp_file() as restore_file:
                data_handler.restore_input_from_recording(prepared_input, [], {'file_name': restore_file})
                self.assertEqual(read_from_file(restore_file), read_from_file(file_name))


class TestOutputInterceptionFileDataHandler(unittest.TestCase):

    def test_prepare_and_restore(self):
        data_handler = OutputInterceptionFileDataHandler(file_path_arg_index=0, file_path_arg_name='file_name')
        with temp_file() as file_name:
            write_to_file(file_name, 'some random meaningless content')
            prepared_input = data_handler.prepare_output_for_recording('key', [file_name], {})
            self.assertEqual(file_name, prepared_input['file_path'])

            with temp_file() as restore_file:
                file_holder = data_handler.restore_output_from_recording(prepared_input)
                file_holder.to_file(restore_file)
                self.assertEqual(read_from_file(restore_file), read_from_file(file_name))
                self.assertEqual(file_holder.file_content, read_from_file(file_name))
                self.assertEqual(file_name, file_holder.output_file_path)

        with temp_file() as file_name:
            write_to_file(file_name, 'some random meaningless content')
            prepared_input = data_handler.prepare_output_for_recording('key', [], {'file_name': file_name})
            self.assertEqual(file_name, prepared_input['file_path'])

            with temp_file() as restore_file:
                file_holder = data_handler.restore_output_from_recording(prepared_input)
                file_holder.to_file(restore_file)
                self.assertEqual(read_from_file(restore_file), read_from_file(file_name))
                self.assertEqual(file_holder.file_content, read_from_file(file_name))
                self.assertEqual(file_name, file_holder.output_file_path)

    def test_intercept_with_file_limit(self):
        data_handler = OutputInterceptionFileDataHandler(file_path_arg_index=0, file_path_arg_name='file_name',
                                                         intercepted_size_limit=1)
        with temp_file() as file_name:
            # Write 2MB file, while interception limit is 2MB
            write_to_file(file_name, 'a' * 2 * 1024 * 1024)
            prepared_input = data_handler.prepare_output_for_recording('key', [file_name], {})
            self.assertEqual(file_name, prepared_input['file_path'])
            self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, prepared_input['file_content'])

            with temp_file() as restore_file:
                file_holder = data_handler.restore_output_from_recording(prepared_input)
                file_holder.to_file(restore_file)
                self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, read_from_file(restore_file))
                self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, file_holder.file_content)
                self.assertEqual(file_name, file_holder.output_file_path)

    def test_intercept_with_file_limit_with_env_variable(self):
        with temp_env('PLAYBACK_INTERCEPTED_FILE_SIZE_LIMIT', '1'):
            data_handler = OutputInterceptionFileDataHandler(file_path_arg_index=0, file_path_arg_name='file_name')
            with temp_file() as file_name:
                # Write 2MB file, while interception limit is 2MB
                write_to_file(file_name, 'a' * 2 * 1024 * 1024)
                prepared_input = data_handler.prepare_output_for_recording('key', [file_name], {})
                self.assertEqual(file_name, prepared_input['file_path'])
                self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, prepared_input['file_content'])

                with temp_file() as restore_file:
                    file_holder = data_handler.restore_output_from_recording(prepared_input)
                    file_holder.to_file(restore_file)
                    self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, read_from_file(restore_file))
                    self.assertEqual(FileInterception.ABOVE_LIMIT_CONTENT, file_holder.file_content)
                    self.assertEqual(file_name, file_holder.output_file_path)



from playback.exceptions import RecordingKeyError
from playback.recording import Recording
from playback.utils.pickle_copy import pickle_copy


class MemoryRecording(Recording):

    def __init__(self, _id=None, recording_data=None, recording_metadata=None):
        """
        :param _id: Id of the recording
        :type _id: str
        :param recording_data: On fetched recording this should contain the recorded data
        :type recording_data: dict
        :param recording_metadata: On fetched recording this should contain the recorded metadata
        :type recording_metadata: dict
        """
        super(MemoryRecording, self).__init__(_id=_id)
        self.recording_data = recording_data or {}
        self.recording_metadata = recording_metadata or {}

    def _set_data(self, key, value):
        """
        Sets data in the recording
        :param key: data key
        :type key: basestring
        :param value: data value (serializable)
        :type value: Any
        """
        self.recording_data[key] = value

    def get_data(self, key):
        """
        :param key: Data key
        :type key: basestring
        :return: Recorded data under given key
        :rtype: Any
        """
        # The contract of the recording requires the implementation to always return a fresh copy of the data.
        # It prevents in place modifications that may be done by the calling code to influence the outcome
        # of the recording playback. That's why we copy here.
        return pickle_copy(self.get_data_direct(key))

    def get_data_direct(self, key):
        """
        :param key: Data key
        :type key: basestring
        :return: Recorded data under given key
        :rtype: Any
        """
        if key not in self.recording_data:
            raise RecordingKeyError(u'Key \'{}\' not found in recording'.format(key).encode("utf-8"))

        return self.recording_data.get(key)

    def get_all_keys(self):
        """
        :return: All recorded keys
        :rtype: list of basestring
        """
        return self.recording_data.keys()

    def _add_metadata(self, metadata):
        """
        :param metadata: Metadata to add to the recording
        :type metadata: dict
        """
        self.recording_metadata.update(metadata)

    def get_metadata(self):
        """
        :return: Recorded metadata
        :rtype: dict
        """
        return self.recording_metadata



# p3ready
from __future__ import absolute_import

import unittest

from playback.tape_cassette import TapeCassette


class TestTapeCassette(unittest.TestCase):

    def test_metadata_content_filter_by_number_value(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key1': 5}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_number_value_not_found(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key1': 8}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_string_value(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key2': "bla"}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_string_value_not_found(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key2': "bla1"}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_list_value(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key2': ["bla", "bla2"]}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla2"}
        filter_metadata = {'key2': ["bla", "bla2"]}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_list_operator_object_value(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key2': ["bla4", "bla5", {'operator': '=', 'value': "bla"}]}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_list_value_not_found(self):
        recording_metadata = {'key1': 5, 'key2': "bla"}
        filter_metadata = {'key2': ["bla1", "bla2"]}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla2"}
        filter_metadata = {'key2': ["bla", "bla3"]}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_object_value(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'obj': {'key1': 6, 'key2': "bla"}}
        filter_metadata = {'obj': {'key1': 6, 'key2': "bla"}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        filter_metadata = {'obj': {'key1': 6, 'key2': "bla1"}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_operator_object_value_equal(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5}
        filter_metadata = {'duration': {'operator': '=', 'value': 5}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 6}
        filter_metadata = {'duration': {'operator': '=', 'value': 5}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_operator_object_value_less(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 4.9}
        filter_metadata = {'duration': {'operator': '<', 'value': 5}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5}
        filter_metadata = {'duration': {'operator': '<', 'value': 5}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_operator_object_value_less_equal(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5}
        filter_metadata = {'duration': {'operator': '<=', 'value': 5}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5.1}
        filter_metadata = {'duration': {'operator': '<=', 'value': 5}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_operator_object_value_greater(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5.1}
        filter_metadata = {'duration': {'operator': '>', 'value': 5}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5}
        filter_metadata = {'duration': {'operator': '>', 'value': 5}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

    def test_metadata_content_filter_by_operator_object_value_greater_equal(self):
        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 5.1}
        filter_metadata = {'duration': {'operator': '>=', 'value': 5}}
        self.assertTrue(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))

        recording_metadata = {'key1': 5, 'key2': "bla", 'duration': 4.9}
        filter_metadata = {'duration': {'operator': '>=', 'value': 5}}
        self.assertFalse(TapeCassette.match_against_recorded_metadata(filter_metadata, recording_metadata))



from abc import ABCMeta, abstractmethod
from fnmatch import fnmatch


class TapeCassette(object):
    """
    An abstract class that acts as a storage driver for TapeRecorder to store and fetch recordings
    """

    __metaclass__ = ABCMeta

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @abstractmethod
    def get_recording(self, recording_id):
        """
        Get recording stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: playback.recording.Recording
        :raises: playback.exceptions.NoSuchRecording
        """
        pass

    def get_recording_metadata(self, recording_id):
        """
        Get recording's metadata stored with the given id
        :param recording_id: If of recording to fetch
        :type recording_id: basestring
        :return: Recording of the given id
        :rtype: dict
        :raises: playback.exceptions.NoSuchRecording
        """
        return self.get_recording(recording_id).get_metadata()

    @abstractmethod
    def create_new_recording(self, category):
        """
        Creates a new recording object
        :param category: A category to classify the recording in (e.g operation class) (serializable)
        :type category: Any
        :return: Creates a new recording object
        :rtype: playback.recording.Recording
        """
        pass

    def abort_recording(self, recording=None):
        # pylint: disable=no-self-use
        """
        Aborts given recording without saving it
        :param recording: Recording to abort
        :type recording: playback.recording.Recording
        """
        recording.close()

    def save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        self._save_recording(recording)
        recording.close()

    @abstractmethod
    def _save_recording(self, recording):
        """
        Saves given recording
        :param recording: Recording to save
        :type recording: playback.recording.Recording
        """
        pass

    @abstractmethod
    def iter_recording_ids(self, category, start_date=None, end_date=None, metadata=None, limit=None,
                           random_results=False):
        """
        Creates an iterator of recording ids matching the given search parameters
        :param category: Recordings category
        :type category: str
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :param random_results: True to return result in random order
        :type random_results: bool
        :return: Iterator of recording ids matching the given parameters
        :rtype: collections.Iterator[str]
        """
        pass

    def iter_recordings_metadata(self, category, start_date=None, end_date=None, metadata=None, limit=None):
        """
        Creates an iterator of recordings metadata matching the given search parameters
        :param category: Recordings category
        :type category: str
        :param start_date: Optional recording start date (need to be given in utc time)
        :type start_date: datetime.datetime
        :param end_date: Optional recording end date (need to be given in utc time)
        :type end_date: datetime.datetime
        :param metadata: Optional metadata values to filter by
        :type metadata: dict
        :param limit: Optional limit on number of ids to fetch
        :type limit: int
        :return: Iterator of recording ids matching the given parameters
        :rtype: collections.Iterator[dict]
        """
        for recording_id in self.iter_recording_ids(
                category, start_date, end_date, metadata, limit):
            yield self.get_recording_metadata(recording_id)

    @staticmethod
    def match_against_recorded_metadata(filter_by_metadata, recording_metadata):
        """
        :param filter_by_metadata: Metadata to match against
        :type filter_by_metadata: dict
        :param recording_metadata: Metadata of a recording
        :type recording_metadata: dict
        :return: Whether the recorded metadata matches the filter metadata
        :rtype: bool
        """
        for k, v in filter_by_metadata.items():  # pylint: disable=invalid-name
            recorded_value = recording_metadata.get(k)
            if not TapeCassette._match_metadata_value(v, recorded_value):
                return False

        return True

    @staticmethod
    def _match_metadata_value(match_value, recorded_value):
        """
        :param match_value: metadata value to match against
        :param recorded_value: metadata value in the recording
        :return: If the recorded value matches the match value
        :rtype: bool
        """
        if isinstance(match_value, list):
            return any(TapeCassette._match_metadata_value(value, recorded_value) for value in match_value)

        if isinstance(match_value, dict) and 'operator' in match_value and 'value' in match_value:
            return TapeCassette._operator_filter(recorded_value, match_value)

        if recorded_value is None and match_value is not None:
            return False

        if isinstance(match_value, str):
            return fnmatch(recorded_value, match_value)

        return recorded_value == match_value

    @staticmethod
    def _operator_filter(recorded_value, metadata_value):
        """
        Check if this is an operator metadata filter and its value is in range
        """
        result = False
        if metadata_value['operator'] == '=':
            result = recorded_value == metadata_value['value']
        if metadata_value['operator'] == '<':
            result = recorded_value < metadata_value['value']
        if metadata_value['operator'] == '<=':
            result = recorded_value <= metadata_value['value']
        if metadata_value['operator'] == '>':
            result = recorded_value > metadata_value['value']
        if metadata_value['operator'] == '>=':
            result = recorded_value >= metadata_value['value']

        return result

    @abstractmethod
    def extract_recording_category(self, recording_id):
        """
        :param recording_id: Recording id to extract category from
        :type recording_id: str
        :return: Recording's category
        :rtype: str
        """
        pass

    def close(self):
        """
        Close this cassette and release any underlying resources
        """
        pass



import logging
from collections import defaultdict, OrderedDict
from datetime import datetime, timedelta
from playback.studio.equalizer import Equalizer
from playback.studio.recordings_lookup import find_matching_recording_ids, RecordingLookupProperties

_logger = logging.getLogger(__name__)


class PlaybackStudio(object):
    """
    Studio that runs multiple playbacks for multiple category using the equalizer to run comparison
    """
    DEFAULT_LOOKUP_PROPERTIES = RecordingLookupProperties(
        start_date=datetime.utcnow() - timedelta(days=7), limit=20)

    def __init__(self, categories, equalizer_tuner, tape_recorder, lookup_properties=None, recording_ids=None,
                 compare_execution_config=None):
        """
        :param categories: Categories (operations) to conduct comparison for
        :type categories: list of str
        :param recording_ids: List of specific recording ids to play, when given categories are ignored
        :type recording_ids: list of str
        :param equalizer_tuner: Given a category return a corresponding equalizer tuning to be used for playback and
        comparison
        :type equalizer_tuner: playback.studio.equalizer_tuning.EqualizerTuner
        :param lookup_properties: Lookup properties to use for all recordings
        :type lookup_properties: RecordingLookupProperties
        :param tape_recorder: The tape recorder that will be used to play the recordings
        :type tape_recorder: playback.tape_recorder.TapeRecorder
        :param compare_execution_config: Configuration specific to the comparison execution flow
        :type compare_execution_config: CompareExecutionConfig
        """
        self.categories = categories
        self.recording_ids = recording_ids
        self.equalizer_tuner = equalizer_tuner
        self.lookup_properties = lookup_properties or self.DEFAULT_LOOKUP_PROPERTIES
        self.tape_recorder = tape_recorder
        self.compare_execution_config = compare_execution_config

    def play(self):
        """
        Fetch and play recording of all categories and run comparison on each one
        :return: Comparison per category of all playbacks
        :rtype: dict[(str, collections.Iterator[(playback.studio.equalizer.Comparison or Exception)]]
        """
        if self.recording_ids:
            categories_recordings = self._group_recording_ids_by_categories()
        else:
            categories_recordings = {c: None for c in self.categories}

        result = {}
        for category, recording_ids in categories_recordings.items():
            result[category] = self._play_category(category, recording_ids)
        return result

    def _group_recording_ids_by_categories(self):
        """
        :return: Recording ids groups by categories
        :rtype: dict[str, list of str]
        """
        grouping = defaultdict(list)
        for recording_id in self.recording_ids:
            category = self.tape_recorder.tape_cassette.extract_recording_category(recording_id)
            grouping[category].append(recording_id)
        # We want deterministic order of playback
        return OrderedDict(sorted(grouping.items()))

    def _play_category(self, category, recording_ids):
        """
        Play and compare recordings of a single category
        :param category: Category to play
        :type category: str
        :param recording_ids: List of specific recording ids to play, None means to fetch recordings using the
        lookup parameters
        :type recording_ids: None or list of str
        :return: Comparison of all playback of current category
        :rtype: (list of playback.studio.equalizer.Comparison) or Exception
        """
        _logger.info(u'Playing Category {}'.format(category))
        try:
            tuning = self.equalizer_tuner.create_category_tuning(category)
        except Exception as ex:  # pylint: disable=broad-except
            _logger.info(u'Cannot tune equalizer for category {} - {}'.format(category, ex))
            return ex

        if recording_ids:
            recording_id_iterator = iter(recording_ids)
        else:
            recording_id_iterator = find_matching_recording_ids(
                self.tape_recorder, category, self.lookup_properties)

        def player(recording_id):
            return self.tape_recorder.play(recording_id, tuning.playback_function)

        equalizer = Equalizer(recording_id_iterator, player, tuning.result_extractor, tuning.comparator,
                              comparison_data_extractor=tuning.comparison_data_extractor,
                              compare_execution_config=self.compare_execution_config)
        return equalizer.run_comparison()



class TapeRecorderException(Exception):
    """
    Base exception for tape recorder
    """
    pass


class InputInterceptionKeyCreationError(TapeRecorderException):
    """
    Exception while creating input interception key
    """
    pass


class RecordingKeyError(TapeRecorderException):
    """
    Exception when key is not found in a recording
    """
    pass


class OperationExceptionDuringPlayback(TapeRecorderException):
    """
    Exception was caught when running the function during a playback
    """
    pass


class NoSuchRecording(TapeRecorderException):
    """
    Recording with given key is not found
    """
    def __init__(self, recording_id):
        """
        :param recording_id: Missing recording id
        :type recording_id: basestring
        """
        super(NoSuchRecording, self).__init__(('No such recording: ' + recording_id).encode('utf-8'))



from examples.flask.playback_context import tape_recorder, init_playback_mode
from playback.studio.equalizer import ComparatorResult, EqualityStatus
from playback.studio.equalizer_tuning import EqualizerTuner, EqualizerTuning
from playback.studio.studio import PlaybackStudio
from playback.tape_recorder import TapeRecorder


class ContentBasedEqualizerTune(EqualizerTuner):

    @staticmethod
    def playback_function(recording):
        """
        Rerun the operation given the recording
        """
        # The class it self is saved as metadata, we instantiate it and trigger the post method
        operation_class = recording.get_metadata()[TapeRecorder.OPERATION_CLASS]()
        operation_class.post()

    @staticmethod
    def result_extractor(outputs):
        """
        For this service we consider the return result of the operation as its output
        """
        return next(o.value['args'][0] for o in outputs if TapeRecorder.OPERATION_OUTPUT_ALIAS in o.key)

    @staticmethod
    def comparator(recorded, played_back):
        """
        Compare recorded value vs the playback back value, doing a basic equality check
        """
        equal = recorded == played_back
        return ComparatorResult(EqualityStatus.Equal if equal else EqualityStatus.Different,
                                message="recorded: {}, played_back: {}".format(recorded, played_back))

    def create_category_tuning(self, category):
        """
        Create a tuning with the needed plugins to rerun the operation and compare the results
        """
        return EqualizerTuning(self.playback_function, self.result_extractor, self.comparator)


init_playback_mode()


studio = PlaybackStudio(categories=['ContentLengthEndpoint', 'ContentFirstCharsEndpoint'],
                        equalizer_tuner=ContentBasedEqualizerTune(),
                        tape_recorder=tape_recorder)

# Run on all categories recording, compare them and output the result
for category, category_comparisons in studio.play().items():
    print('Category {}'.format(category))
    for comparison_result in category_comparisons:
        print('{recording_id}: {result}'.format(
            recording_id=comparison_result.recording_id,
            result=comparison_result.comparator_status))
    print('')



from jsonpickle import encode, decode


def pickle_copy(value):
    """ copies any object (deeply) by pickly encoding/decoding it
    :type value: any
    :param value: the value you to be copied
    :rtype: any
    """
    return decode(encode(value, unpicklable=True))



from time import time


class Timed(object):
    """
    Tiny timing util that execution time of the scope it wraps
    """
    def __init__(self):
        self.duration = None
        self._start = None

    def __enter__(self):
        self._start = time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.duration = time() - self._start



# p3ready
from __future__ import absolute_import
import unittest

from playback.studio.equalizer import ComparatorResult, EqualityStatus, CompareExecutionConfig
from playback.studio.studio import PlaybackStudio, RecordingLookupProperties
from playback.studio.equalizer_tuning import EqualizerTuning, EqualizerTuner
from playback.tape_recorder import TapeRecorder
from playback.tape_cassettes.in_memory.in_memory_tape_cassette import InMemoryTapeCassette


class TestPlaybackStudio(unittest.TestCase):

    def setUp(self):
        self.tape_cassette = InMemoryTapeCassette()
        self.tape_recorder = TapeRecorder(self.tape_cassette)
        self.tape_recorder.enable_recording()

    def tearDown(self):
        self.tape_cassette.close()

    def test_run_two_categories(self):
        self._test_run_two_operations(use_recording_ids=False)

    def test_run_specific_recording_ids(self):
        self._test_run_two_operations(use_recording_ids=True)

    def _test_run_two_operations(self, use_recording_ids):
        class A(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 'AAA'

        class B(object):

            @self.tape_recorder.operation()
            def execute(self):
                return 'BBB'

        A().execute()
        A().execute()
        B().execute()

        if not use_recording_ids:
            categories = ['A', 'B', 'C']
            recording_ids = None
        else:
            categories = None
            recording_ids = self.tape_cassette.get_all_recording_ids()

        class MockEqualizerTuner(EqualizerTuner):

            def create_category_tuning(self, category):
                if category not in ['A', 'B']:
                    raise Exception("Unsupported")

                def result_extractor(outputs):
                    return next(o.value['args'][0] for o in outputs if TapeRecorder.OPERATION_OUTPUT_ALIAS in o.key)

                def comparator(expected, actual, message):
                    return ComparatorResult(EqualityStatus.Equal if expected == actual else EqualityStatus.Different,
                                            message)

                def comparison_data_extractor(recording):
                    return {'message': category}

                def playback_function(recording):
                    cls_name = recording.get_metadata()[TapeRecorder.OPERATION_CLASS]
                    if 'studio.A' in list(cls_name.values())[0]:
                        return A().execute()
                    return B().execute()

                return EqualizerTuning(
                    playback_function=playback_function, result_extractor=result_extractor, comparator=comparator,
                    comparison_data_extractor=comparison_data_extractor)

        equalizer_tuner = MockEqualizerTuner()
        start_date = 'a'
        studio = PlaybackStudio(categories, equalizer_tuner, self.tape_recorder,
                                lookup_properties=RecordingLookupProperties(start_date), recording_ids=recording_ids,
                                compare_execution_config=CompareExecutionConfig(keep_results_in_comparison=True))
        result = studio.play()

        a_results = list(result['A'])
        b_results = list(result['B'])

        self.assertEqual(2, len(a_results))
        self.assertTrue(all('AAA' == result.actual for result in a_results))
        self.assertTrue(all('AAA' == result.expected for result in a_results))
        self.assertTrue(all(EqualityStatus.Equal == result.comparator_status.equality_status for result in a_results))
        self.assertTrue(all('A' == result.comparator_status.message for result in a_results))

        self.assertEqual(1, len(b_results))
        self.assertTrue(all('BBB' == result.actual for result in b_results))
        self.assertTrue(all('BBB' == result.expected for result in b_results))
        self.assertTrue(all(EqualityStatus.Equal == result.comparator_status.equality_status for result in b_results))
        self.assertTrue(all('B' == result.comparator_status.message for result in b_results))



import os
import sys
import setuptools
from setuptools.command.install import install

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = '0.3.21'


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


setuptools.setup(
    name="playback-studio",
    version=VERSION,
    author="Optibus",
    author_email="eitan@optibus.com",
    description="Record your service operations in production and replay them locally at any time in a sandbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Optibus/playback",
    packages=setuptools.find_packages(),
    install_requires=[
        'parse==1.6.6',
        'jsonpickle==0.9.3',
        'six>=1.15.0',
        'contextlib2==0.6.0',
        'decorator==4.4.2'
    ],
    extras_require={'dev': [
        'mock==2.0.0',
        'rsa<=4.0; python_version < \'3\'',
        'python-jose<=0.3.2; python_version < \'3\'',
        'moto==1.3.13',
        'pytest==4.6.9',
        'parameterized==0.7.0',
        'Werkzeug==0.16.1',
        'Flask==1.1.2',
        'flask-restplus==0.13.0',
        'future==0.18.2',
        'pylint==2.6.0; python_version > "3.0"',
        'pylint-junit==0.3.2; python_version > "3.0"',
        'flake8==3.8.4; python_version > "3.0"',
        'flake8-formatter-junit-xml==0.0.6; python_version > "3.0"',
        'pytest-cov==2.10.1; python_version > "3.0"'
    ]},
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=2.7',
    cmdclass={
        'verify': VerifyVersionCommand,
    },
)



import unittest

from playback.tape_cassettes.asynchronous.async_record_only_tape_cassette import AsyncRecordOnlyTapeCassette
from tests.mocks.delayed_in_memory_tape_cassette import DelayedInMemoryTapeCassette
from playback.utils.timing_utils import Timed


class TestAsyncTapeCassette(unittest.TestCase):

    def test_async_set_data_behavior(self):
        self._test_async_data_set_behavior(add_error_op=False)

    def test_async_set_data_behavior_with_error_op(self):
        self._test_async_data_set_behavior(add_error_op=True)

    def _test_async_data_set_behavior(self, add_error_op):
        in_memory_cassette = DelayedInMemoryTapeCassette(delay=0.1)
        tape_cassette = AsyncRecordOnlyTapeCassette(in_memory_cassette, timeout_on_close=5)
        tape_cassette.start()
        recording = tape_cassette.create_new_recording('category')
        self.assertEqual(recording.id, recording.wrapped_recording.id)

        def error_operation():
            raise Exception('mock')

        with Timed() as timed:
            recording.add_metadata({'c': 3, 'd': 4})
            recording.set_data('a', 2)
            recording.set_data('b', 1)

            if add_error_op:
                tape_cassette._add_async_operation(error_operation)

            tape_cassette.save_recording(recording)
        self.assertLess(timed.duration, 0.1)
        tape_cassette.close()
        _id = in_memory_cassette.get_last_recording_id()
        recording = in_memory_cassette.get_recording(_id)
        self.assertEqual(2, recording.get_data('a'))
        self.assertEqual(1, recording.get_data('b'))
        self.assertEqual({'c': 3, 'd': 4}, recording.get_metadata())

    def test_fetch_recording_operation_raising_errors_read_live_recording_works(self):
        in_memory_cassette = DelayedInMemoryTapeCassette(delay=0.1)
        tape_cassette = AsyncRecordOnlyTapeCassette(in_memory_cassette, timeout_on_close=5)
        tape_cassette.start()
        try:
            with self.assertRaises(TypeError):
                tape_cassette.get_recording('a')
            with self.assertRaises(TypeError):
                tape_cassette.iter_recording_ids('a')
            with self.assertRaises(TypeError):
                tape_cassette.extract_recording_category('a')

            recording = tape_cassette.create_new_recording('category')
            recording.set_data('a', 'a')
            recording.set_data('b', 'b')
            recording.add_metadata({'c': 'c'})
            self.assertEqual('a', recording.get_data('a'))
            self.assertSequenceEqual(['a', 'b'], list(recording.get_all_keys()))
            self.assertEqual({'c': 'c'}, recording.get_metadata())
        finally:
            tape_cassette.close()

    def test_close_without_open_no_error(self):
        in_memory_cassette = DelayedInMemoryTapeCassette(delay=0.1)
        tape_cassette = AsyncRecordOnlyTapeCassette(in_memory_cassette, timeout_on_close=5)
        tape_cassette.close()
